from __future__ import annotations

import shutil
from datetime import datetime, timezone
from typing import TYPE_CHECKING

import pytest

from dissect.target.exceptions import RegistryKeyNotFoundError, UnsupportedPluginError
from dissect.target.helpers.regutil import VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.log import evt, evtx
from dissect.target.plugins.scrape import scrape
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.filesystem import VirtualFilesystem


def mock_registry_log_location(target_win: Target, reg_key_name: str, mock_log_path: str) -> None:
    # Mock eventlog registry key in a specific control set,
    # CurrentControlSet used in preconfigured value is a soft link
    registry_hive = target_win.registry._root

    eventlog_key = evt.WindowsEventlogsMixin.EVENTLOG_REGISTRY_KEY.replace("CurrentControlSet", "ControlSet001")

    try:
        registry_hive.key(eventlog_key)
    except RegistryKeyNotFoundError:
        registry_hive.map_key(eventlog_key, VirtualKey(registry_hive, eventlog_key))

    # Set 'File' registry value to eventlog file path
    registry_key = f"{eventlog_key}\\{reg_key_name}"
    registry_subkey = VirtualKey(registry_hive, registry_key)
    registry_hive.map_key(registry_key, registry_subkey)

    eventlog_security_value = VirtualValue(registry_hive, "File", mock_log_path)
    registry_hive.map_value(registry_key, "File", eventlog_security_value)


@pytest.mark.parametrize(
    ("is_in_directory", "is_in_registry", "duplicate"),
    [
        (True, False, False),
        (False, True, False),
        (True, True, False),
        (True, True, True),
    ],
)
def test_evtx_plugin(
    target_win: Target,
    fs_win: VirtualFilesystem,
    tmp_path: Path,
    is_in_directory: bool,
    is_in_registry: bool,
    duplicate: bool,
) -> None:
    with pytest.raises(UnsupportedPluginError):
        target_win.add_plugin(evtx.EvtxPlugin)

    # Map default log location to pass EvtxPlugin's compatibility check
    fs_win.map_dir("windows/system32/winevt/logs", tmp_path)
    target_win.add_plugin(evtx.EvtxPlugin)

    evtx_log_file = absolute_path("_data/plugins/os/windows/log/evtx/TestLogX.evtx")
    expected_records = 0

    if is_in_directory:
        evtx_dir_file = tmp_path / "TestLogXDir.evtx"
        shutil.copyfile(evtx_log_file, evtx_dir_file)

        # Mock log file in a default directory
        fs_win.map_file("windows/system32/winevt/logs/TestLogMock.evtx", evtx_dir_file)
        expected_records += 5

    if is_in_registry:
        evtx_reg_file = tmp_path / "TestLogXReg.evtx"
        shutil.copyfile(evtx_log_file, evtx_reg_file)

        # Set a log path in the registry key and map that path
        mock_log_path = "somefolder/TestLogMock.EVTX"
        mock_registry_log_location(target_win, "Security", f"C:\\{mock_log_path}")
        fs_win.map_file(mock_log_path, evtx_reg_file)

        # Set path that should be skipped
        mock_registry_log_location(target_win, "MockApp", "C:\\somefolder\\BadFile.wrong")

        # Set path that doesn't exist
        mock_registry_log_location(target_win, "MockApp", "C:\\somefolder\\BadFile.evtx")

        expected_records += 5

    if duplicate:
        # Add a registry path that we should already pick up from the default directory
        # The only difference is that the directory file is mapped to sysvol, whereas we will reference it from C here
        # Expected records shouldn't change
        mock_registry_log_location(target_win, "Security Dupe", "c:/windows/system32/winevt/logs/TestLogMock.evtx")

    records = list(target_win.evtx())

    assert len(records) == expected_records

    # verify that out of all received records, only 5 are unique
    assert len({str(rec.EventID) for rec in records}) == 5


def test_evtx_scraping(target_win: Target) -> None:
    target_win.add_plugin(scrape.ScrapePlugin)

    plugin = evtx.EvtxPlugin(target_win)
    evtx_log_file = absolute_path("_data/plugins/os/windows/log/evtx/TestLogX.evtx")

    with evtx_log_file.open("rb") as fh:
        target_win.disks.add(fh)
        scraped_records = list(plugin.scraped_evtx())

    assert len(scraped_records) == 5


def test_evtx_normalize_values(target_win: Target, fs_win: VirtualFilesystem) -> None:
    """Test if we normalize certain evtx fields correctly."""

    # Example Security.evtx originates from Windows 10 22H2 Pro build 19045.2006,
    # events exported after clean virtual machine post-install.
    security_evtx = absolute_path("_data/plugins/os/windows/log/evtx/Security.evtx")
    fs_win.map_file("Windows\\System32\\winevt\\Logs\\Security.evtx", security_evtx)

    target_win.add_plugin(evtx.EvtxPlugin)

    records = sorted(target_win.evtx(), key=lambda r: r.ts)

    # Verified amount of events in Event Viewer
    assert len(records) == 759

    assert records[0].ts == datetime(2025, 3, 4, 10, 27, 58, 245262, tzinfo=timezone.utc)
    assert records[0].Provider_Name == "Microsoft-Windows-Security-Auditing"
    assert records[0].EventID == 4616
    assert records[0].Channel == "Security"
    assert records[0].Computer == "DESKTOP-L7A1DDP"
    assert records[0].Correlation_ActivityID is None
    assert records[0].SubjectUserSid == "S-1-5-18"
    assert records[0].source == "c:\\Windows\\system32\\winevt\\logs\\Security.evtx"

    assert records[69].ts == datetime(2025, 3, 4, 10, 28, 2, 905350, tzinfo=timezone.utc)
    assert records[69].PrivilegeList is None
    assert records[69].SidHistory is None

    # Should contain None instead of string "-"
    assert {getattr(r, "IpAddress", None) for r in records} == {None, "127.0.0.1"}


@pytest.mark.parametrize(
    ("key", "keys", "expected_key"),
    [
        ("source", {"source", "_target"}, "source_2_duplicate"),
        ("source", {"source", "source_2_duplicate", "_target"}, "source_3_duplicate"),
        ("source", {"source", "source_2", "_target"}, "source_2_duplicate"),
    ],
)
def test_evtx_key_deduplication(key: str, keys: set[str], expected_key: str) -> None:
    """Test if ``unique_keys`` correctly deduplicates key values."""

    assert evtx.unique_key(key, keys) == expected_key


def test_evtx_direct_mode() -> None:
    data_path = absolute_path("_data/plugins/os/windows/log/evtx/TestLogX.evtx")

    target = Target.open_direct([data_path])
    records = list(target.evtx())

    assert len(records) == 5


def test_evtx_build_record_illegal_characters(target_win: Target) -> None:
    """Test that we correctly replace illegal characters in record field names."""
    evtx_record = {
        "Provider_Name": "Microsoft-Windows-Kernel-Boot",
        "Provider_Guid": "{FF44CA15-7A4D-AA4B-BBA5-0998955E531E}",
        "EventID": 85,
        "pSubStatus->PrimaryBlob():.<#/Status": 1,
    }

    record = evtx.EvtxPlugin(target_win)._build_record(
        evtx_record,
        target_win.fs.path("C:/WINDOWS/system32/winevt/logs/Microsoft-Windows-Kernel-Boot%4Operational.evtx"),
    )

    assert record.pSubStatus__PrimaryBlob_______Status == "1"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.linux._os import LinuxPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class SuSEPlugin(LinuxPlugin):
    def __init__(self, target: Target):
        super().__init__(target)

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            # modern and old package managers of suse distros
            if len(list(fs.glob("/etc/*YaST*"))) > 0 or fs.exists("/etc/zypp"):
                return fs
        return None



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.shellitem.lnk import Lnk
from dissect.util import ts

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target

LnkRecord = TargetRecordDescriptor(
    "windows/filesystem/lnk",
    [
        ("path", "lnk_path"),
        ("string", "lnk_name"),
        ("datetime", "lnk_mtime"),
        ("datetime", "lnk_atime"),
        ("datetime", "lnk_ctime"),
        ("path", "lnk_relativepath"),
        ("path", "lnk_workdir"),
        ("string", "lnk_arguments"),
        ("path", "lnk_iconlocation"),
        ("string", "local_base_path"),
        ("string", "common_path_suffix"),
        ("string", "lnk_net_name"),
        ("string", "lnk_device_name"),
        ("path", "lnk_full_path"),
        ("string", "machine_id"),
        ("datetime", "target_mtime"),
        ("datetime", "target_atime"),
        ("datetime", "target_ctime"),
    ],
)


def parse_lnk_file(target: Target, lnk_file: Lnk, lnk_path: TargetPath) -> LnkRecord:
    # we need to get the active codepage from the system to properly decode some values
    codepage = target.codepage or "ascii"

    lnk_net_name = lnk_device_name = None

    if lnk_file.link_header:
        lnk_name = lnk_file.stringdata.name_string.string if lnk_file.flag("has_name") else None

        lnk_mtime = ts.from_unix(lnk_path.stat().st_mtime)
        lnk_atime = ts.from_unix(lnk_path.stat().st_atime)
        lnk_ctime = ts.from_unix(lnk_path.stat().st_ctime)

        lnk_relativepath = (
            target.fs.path(lnk_file.stringdata.relative_path.string) if lnk_file.flag("has_relative_path") else None
        )
        lnk_workdir = (
            target.fs.path(lnk_file.stringdata.working_dir.string) if lnk_file.flag("has_working_dir") else None
        )
        lnk_iconlocation = (
            target.fs.path(lnk_file.stringdata.icon_location.string) if lnk_file.flag("has_icon_location") else None
        )
        lnk_arguments = lnk_file.stringdata.command_line_arguments.string if lnk_file.flag("has_arguments") else None
        local_base_path = (
            lnk_file.linkinfo.local_base_path.decode(codepage)
            if lnk_file.flag("has_link_info") and lnk_file.linkinfo.flag("volumeid_and_local_basepath")
            else None
        )
        common_path_suffix = (
            lnk_file.linkinfo.common_path_suffix.decode(codepage) if lnk_file.flag("has_link_info") else None
        )

        if local_base_path and common_path_suffix:
            lnk_full_path = target.fs.path(local_base_path + common_path_suffix)
        elif local_base_path and not common_path_suffix:
            lnk_full_path = target.fs.path(local_base_path)
        else:
            lnk_full_path = None

        if lnk_file.flag("has_link_info") and lnk_file.linkinfo.flag("common_network_relative_link_and_pathsuffix"):
            lnk_net_name = (
                lnk_file.linkinfo.common_network_relative_link.net_name.decode()
                if lnk_file.linkinfo.common_network_relative_link.net_name
                else None
            )
            lnk_device_name = (
                lnk_file.linkinfo.common_network_relative_link.device_name.decode()
                if lnk_file.linkinfo.common_network_relative_link.device_name
                else None
            )
        try:
            machine_id = lnk_file.extradata.TRACKER_PROPS.machine_id.decode(codepage).strip("\x00")
        except AttributeError:
            machine_id = None

        target_mtime = ts.wintimestamp(lnk_file.link_header.write_time)
        target_atime = ts.wintimestamp(lnk_file.link_header.access_time)
        target_ctime = ts.wintimestamp(lnk_file.link_header.creation_time)

        return LnkRecord(
            lnk_path=lnk_path,
            lnk_name=lnk_name,
            lnk_mtime=lnk_mtime,
            lnk_atime=lnk_atime,
            lnk_ctime=lnk_ctime,
            lnk_relativepath=lnk_relativepath,
            lnk_workdir=lnk_workdir,
            lnk_iconlocation=lnk_iconlocation,
            lnk_arguments=lnk_arguments,
            local_base_path=local_base_path,
            common_path_suffix=common_path_suffix,
            lnk_full_path=lnk_full_path,
            lnk_net_name=lnk_net_name,
            lnk_device_name=lnk_device_name,
            machine_id=machine_id,
            target_mtime=target_mtime,
            target_atime=target_atime,
            target_ctime=target_ctime,
            _target=target,
        )
    return None


class LnkPlugin(Plugin):
    """Windows lnk plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.folders = ["programdata", "users", "windows"]

    def check_compatible(self) -> None:
        for folder in self.folders:
            if self.target.fs.path(f"sysvol/{folder}").exists():
                return
        raise UnsupportedPluginError("No folders containing link files found")

    @arg("-p", "--path", help="path to directory or .lnk file in target")
    @export(record=LnkRecord)
    def lnk(self, path: str | None = None) -> Iterator[LnkRecord]:
        """Parse all .lnk files in /ProgramData, /Users, and /Windows or from a specified path in record format.

        Yields a LnkRecord record with the following fields:

        .. code-block:: text

            lnk_path (path): Path of the link (.lnk) file.
            lnk_name (string): Name of the link (.lnk) file.
            lnk_mtime (datetime): Modification time of the link (.lnk) file.
            lnk_atime (datetime): Access time of the link (.lnk) file.
            lnk_ctime (datetime): Creation time of the link (.lnk) file.
            lnk_relativepath (path): Relative path of target file to the link (.lnk) file.
            lnk_workdir (path): Path of the working directory the link (.lnk) file will execute from.
            lnk_iconlocation (path): Path of the display icon used for the link (.lnk) file.
            lnk_arguments (string): Command-line arguments passed to the target (linked) file.
            local_base_path (string): Absolute path of the target (linked) file.
            common_path_suffix (string): Suffix of the local_base_path.
            lnk_full_path (string): Full path of the linked file. Made from local_base_path and common_path_suffix.
            lnk_net_name (string): Specifies a server share path; for example, "\\\\server\\share".
            lnk_device_name (string): Specifies a device; for example, the drive letter "D:"
            machine_id (string): The NetBIOS name of the machine where the linked file was last known to reside.
            target_mtime (datetime): Modification time of the target (linked) file.
            target_atime (datetime): Access time of the target (linked) file.
            target_ctime (datetime): Creation time of the target (linked) file.
        """

        for entry in self.lnk_entries(path):
            try:
                lnk_file = Lnk(entry.open())
                if (lnk_rec := parse_lnk_file(self.target, lnk_file, entry)) is not None:
                    yield lnk_rec
            except Exception as e:  # noqa: PERF203
                self.target.log.warning("Failed to parse link file %s", entry)
                self.target.log.debug("", exc_info=e)

    def lnk_entries(self, path: str | None = None) -> Iterator[TargetPath]:
        if path:
            target_path = self.target.fs.path(path)
            if not target_path.exists():
                self.target.log.error("Provided path %s does not exist on target", target_path)
                return

            if target_path.is_file():
                yield target_path
            else:
                yield from target_path.rglob("*.lnk")
        else:
            for folder in self.folders:
                yield from self.target.fs.path("sysvol").joinpath(folder).rglob("*.lnk")



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_users_plugin(target_win_users: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    users = list(target_win_users.users())
    assert len(users) == 2

    users_details = list(target_win_users.user_details.all())
    assert len(users_details) == 2
    assert {d.user.sid for d in users_details} == {u.sid for u in users}

    assert target_win_users.user_details.find(sid="S-1-5-21-3263113198-3007035898-945866154-1002")
    john = target_win_users.user_details.find(username="John")
    assert john

    details = target_win_users.user_details.get(john.user)
    assert details
    assert details.user == john.user
    assert details.home_path == target_win_users.fs.path("C:/Users/John")

    users_with_home = list(target_win_users.user_details.all_with_home())
    assert len(users_with_home) == 0  # no users have home dirs

    fs_win.map_dir("Users\\John", tmp_path)
    users_with_home = list(target_win_users.user_details.all_with_home())
    assert len(users_with_home) == 1  # only John has a home dir

    home_dirs = list(target_win_users.user_details.all_home_paths)
    assert len(home_dirs) == 1
    assert home_dirs[0] == target_win_users.fs.path("C:/Users/John")


def test_misc_users(target_win_users: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    fs_win.map_dir("Windows/SysWOW64/config/systemprofile", tmp_path)
    users_with_home = list(target_win_users.user_details.all_with_home())
    assert len(users_with_home) == 1  # Misc home dir is found
    assert users_with_home[0].user.sid == "S-1-5-18"
    assert users_with_home[0].home_path == target_win_users.fs.path("c:/Windows/SysWOW64/config/systemprofile")

    home_dirs = list(target_win_users.user_details.all_home_paths)
    assert len(home_dirs) == 1  # Misc home dir is found
    assert home_dirs[0] == target_win_users.fs.path("c:/Windows/SysWOW64/config/systemprofile")


def test_users_plugin_find_no_params(target_unix_users: Target) -> None:
    with pytest.raises(ValueError, match="Either sid or uid or username is expected"):
        target_unix_users.user_details.find()


def test_users_plugin_find_no_usser(target_unix_users: Target) -> None:
    user_details = target_unix_users.user_details.find(uid=13)
    assert user_details is None


def test_users_plugin_find_uid0(target_unix_users: Target) -> None:
    user_details = target_unix_users.user_details.find(uid=0)
    assert user_details is not None
    assert user_details.user.uid == 0



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.plugins.child.colima import ColimaChildTargetPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_child_colima(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/root/.colima/default/colima.yaml", BytesIO())
    fs_unix.map_file_fh("/root/.colima/_lima/colima/diffdisk", BytesIO())
    fs_unix.map_file_fh("/root/.colima/test/colima.yaml", BytesIO())
    fs_unix.map_file_fh("/root/.colima/_lima/colima-test/diffdisk", BytesIO())
    fs_unix.map_file_fh("/home/user/.config/colima/default/colima.yaml", BytesIO())
    fs_unix.map_file_fh("/home/user/.config/colima/_lima/colima/diffdisk", BytesIO())

    target_unix_users.add_plugin(ColimaChildTargetPlugin)

    children = list(target_unix_users.list_children())

    assert len(children) == 3

    child_id, child_record = children[0]
    assert child_id == "0"
    assert child_record.type == "colima"
    assert child_record.name == "default"
    assert child_record.path == "/root/.colima/_lima/colima/diffdisk"

    child_id, child_record = children[1]
    assert child_id == "1"
    assert child_record.name == "test"
    assert child_record.path == "/root/.colima/_lima/colima-test/diffdisk"

    child_id, child_record = children[2]
    assert child_id == "2"
    assert child_record.path == "/home/user/.config/colima/_lima/colima/diffdisk"



#!/usr/bin/env python
from __future__ import annotations

import argparse
import os
import random
import sys
from typing import TYPE_CHECKING

from dissect.cstruct import utils

from dissect.target.exceptions import TargetError
from dissect.target.helpers.logging import get_logger
from dissect.target.helpers.scrape import recover_string
from dissect.target.plugins.scrape.qfind import QFindMatchRecord, QFindPlugin
from dissect.target.tools.query import record_output
from dissect.target.tools.utils.cli import (
    catch_sigpipe,
    configure_generic_arguments,
    open_targets,
    process_generic_arguments,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from dissect.target.container import Container
    from dissect.target.target import Target
    from dissect.target.volume import Volume


log = get_logger(__name__)

NO_COLOR = os.getenv("NO_COLOR")
COLOR_GREY = "\033[38;5;248m"


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="Find a needle in a haystack.",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )

    parser.add_argument("targets", metavar="TARGETS", nargs="*", help="Targets to load")
    parser.add_argument(
        "-R", "--raw", action="store_true", help="show raw hex dumps instead of post-processed string output"
    )
    parser.add_argument("--allow-non-ascii", action="store_true", help="allow non-ASCII characters in the output")
    parser.add_argument("-j", "--json", action="store_true", help="output records as json")
    parser.add_argument("-r", "--record", action="store_true", help="output records")
    parser.add_argument("-s", "--strings", action="store_true", help="print record output as string")

    for args, kwargs in getattr(QFindPlugin.qfind, "__args__", []):
        parser.add_argument(*args, **kwargs)

    configure_generic_arguments(parser)

    args, _ = parser.parse_known_args()
    process_generic_arguments(parser, args)

    if not args.targets:
        log.error("No targets provided")
        return 1

    rs = None
    if args.record or args.json:
        rs = record_output(args.strings, args.json)

    try:
        for target in open_targets(args):
            hit: QFindMatchRecord
            for hit in target.qfind(
                args.needles,
                args.needle_file,
                args.encoding,
                args.no_hex_decode,
                args.regex,
                args.ignore_case,
                args.unique,
                args.window,
                args.strip_null_bytes,
                progress=progress_handler(target),
            ):
                if rs:
                    rs.write(hit)
                    continue

                header = f"[{hit.offset:#08x} @ {hit.needle} ({hit.codec})]"

                if not NO_COLOR:
                    header = utils.COLOR_WHITE + header + utils.COLOR_NORMAL

                before_offset = max(0, hit.offset - args.window)
                needle_len = len(hit.match)

                print(f"\r{header}")

                if args.raw:
                    palette = (
                        [(hit.offset - before_offset, utils.COLOR_NORMAL), (needle_len, utils.COLOR_BG_RED)]
                        if not NO_COLOR
                        else None
                    )
                    utils.hexdump(hit.buffer, palette, offset=before_offset)

                else:
                    codec = "utf-8" if hit.codec == "hex" else hit.codec
                    before_part = recover_string(
                        hit.buffer[: hit.offset - before_offset], codec, reverse=True, ascii=not args.allow_non_ascii
                    )
                    after_part = recover_string(
                        hit.buffer[hit.offset - before_offset :], codec, ascii=not args.allow_non_ascii
                    )
                    hit = (
                        before_part,
                        (utils.COLOR_BG_RED if not NO_COLOR else ""),
                        after_part[:needle_len],
                        (utils.COLOR_NORMAL if not NO_COLOR else ""),
                        after_part[needle_len:],
                    )
                    print("".join(hit))

    except TargetError as e:
        log.error(e)  # noqa: TRY400
        log.debug("", exc_info=e)
        return 1

    if not rs:
        print(end="\r\n")

    return 0


def progress_handler(target: Target) -> Callable[[Container | Volume, int, int], None]:
    """Progress handler of the qfind plugin."""
    current_disk = None
    animations = [
        ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"],
        ["⠄", "⠆", "⠇", "⠋", "⠙", "⠸", "⠰", "⠠", "⠰", "⠸", "⠙", "⠋", "⠇", "⠆"],
        [
            "⢀⠀",
            "⡀⠀",
            "⠄⠀",
            "⢂⠀",
            "⡂⠀",
            "⠅⠀",
            "⢃⠀",
            "⡃⠀",
            "⠍⠀",
            "⢋⠀",
            "⡋⠀",
            "⠍⠁",
            "⢋⠁",
            "⡋⠁",
            "⠍⠉",
            "⠋⠉",
            "⠋⠉",
            "⠉⠙",
            "⠉⠙",
            "⠉⠩",
            "⠈⢙",
            "⠈⡙",
            "⢈⠩",
            "⡀⢙",
            "⠄⡙",
            "⢂⠩",
            "⡂⢘",
            "⠅⡘",
            "⢃⠨",
            "⡃⢐",
            "⠍⡐",
            "⢋⠠",
            "⡋⢀",
            "⠍⡁",
            "⢋⠁",
            "⡋⠁",
            "⠍⠉",
            "⠋⠉",
            "⠋⠉",
            "⠉⠙",
            "⠉⠙",
            "⠉⠩",
            "⠈⢙",
            "⠈⡙",
            "⠈⠩",
            "⠀⢙",
            "⠀⡙",
            "⠀⠩",
            "⠀⢘",
            "⠀⡘",
            "⠀⠨",
            "⠀⢐",
            "⠀⡐",
            "⠀⠠",
            "⠀⢀",
            "⠀⡀",
        ],
        [
            "⠁",
            "⠂",
            "⠄",
            "⡀",
            "⡈",
            "⡐",
            "⡠",
            "⣀",
            "⣁",
            "⣂",
            "⣄",
            "⣌",
            "⣔",
            "⣤",
            "⣥",
            "⣦",
            "⣮",
            "⣶",
            "⣷",
            "⣿",
            "⡿",
            "⠿",
            "⢟",
            "⠟",
            "⡛",
            "⠛",
            "⠫",
            "⢋",
            "⠋",
            "⠍",
            "⡉",
            "⠉",
            "⠑",
            "⠡",
            "⢁",
        ],
        [
            "[010010]",
            "[001100]",
            "[100101]",
            "[111010]",
            "[111101]",
            "[010111]",
            "[101011]",
            "[111000]",
            "[110011]",
            "[110101]",
        ],
        ["◐", "◓", "◑", "◒"],
        ["🕛 ", "🕐 ", "🕑 ", "🕒 ", "🕓 ", "🕔 ", "🕕 ", "🕖 ", "🕗 ", "🕘 ", "🕙 ", "🕚 "],
        [
            "▐⠂       ▌",
            "▐⠈       ▌",
            "▐ ⠂      ▌",
            "▐ ⠠      ▌",
            "▐  ⡀     ▌",
            "▐  ⠠     ▌",
            "▐   ⠂    ▌",
            "▐   ⠈    ▌",
            "▐    ⠂   ▌",
            "▐    ⠠   ▌",
            "▐     ⡀  ▌",
            "▐     ⠠  ▌",
            "▐      ⠂ ▌",
            "▐      ⠈ ▌",
            "▐       ⠂▌",
            "▐       ⠠▌",
            "▐       ⡀▌",
            "▐      ⠠ ▌",
            "▐      ⠂ ▌",
            "▐     ⠈  ▌",
            "▐     ⠂  ▌",
            "▐    ⠠   ▌",
            "▐    ⡀   ▌",
            "▐   ⠠    ▌",
            "▐   ⠂    ▌",
            "▐  ⠈     ▌",
            "▐  ⠂     ▌",
            "▐ ⠠      ▌",
            "▐ ⡀      ▌",
            "▐⠠       ▌",
        ],
        [
            " 🤜\u3000\u3000\u3000\u3000🤛 ",
            " 🤜\u3000\u3000\u3000\u3000🤛 ",
            " 🤜\u3000\u3000\u3000\u3000🤛 ",
            " \u3000🤜\u3000\u3000🤛\u3000 ",
            " \u3000\u3000🤜🤛\u3000\u3000 ",
            " \u3000🤜✨🤛\u3000\u3000 ",
            " 🤜\u3000✨\u3000🤛\u3000 ",
        ],
        [
            "😐 ",
            "😐 ",
            "😮 ",
            "😮 ",
            "😦 ",
            "😦 ",
            "😧 ",
            "😧 ",
            "🤯 ",
            "💥 ",
            "✨ ",
            "\u3000 ",
            "\u3000 ",
            "\u3000 ",
        ],
        ["_", "_", "_", "-", "`", "`", "'", "´", "-", "_", "_", "_"],  # noqa: RUF001
    ]
    animation = random.choice(animations)
    char = 0

    def update(disk: Container | Volume, offset: int, size: int) -> None:
        nonlocal current_disk, char

        if current_disk is None:
            sys.stderr.write(f"{utils.COLOR_WHITE}{target}{utils.COLOR_NORMAL}\n")

        if current_disk != disk:
            sys.stderr.write(f"\n{utils.COLOR_WHITE}[Current disk: {disk}]{utils.COLOR_NORMAL}\n")
            current_disk = disk

        sys.stderr.write(f"\r{COLOR_GREY}{offset / float(size) * 100:0.2f}% {animation[char]}{utils.COLOR_NORMAL}")
        sys.stderr.flush()

        if offset % (1337 * 3) == 0:
            char = 0 if char == (len(animation) - 1) else char + 1

    return update


if __name__ == "__main__":
    main()



from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import BinaryIO, TypeAlias

from dissect.target.filesystem import FilesystemEntry
from dissect.target.helpers.magic import mimetypes

MagicSignature: TypeAlias = tuple[str, str, bytes]
MagicResult: TypeAlias = str | None


class Magic:
    """Magic is a helper class for identifying files to a file type by magic bytes.

    This class mimics ``python-magic`` behaviour, however this implementation does
    not depend on the ``libmagic`` C-library system package. Instead a precompiled
    XML document of FreeDesktop's shared-mime-info project is used to identify files.

    Falls back to file extensions (where applicable) if no magic header is detected
    in the provided buffer.

    The underlying implementation currently does not support nested magic definitions
    and logical AND/OR matches, which somewhat limits the supported FreeDesktop XML
    magic signature set.

    .. code-block:: python

        from dissect.target.helpers import magic

        >>> magic.from_buffer(b"SQLite format 3\x00FILE DATA\x4d\x3c\xb2\xa1")
        "SQLite3 database"

        >>> magic.from_file(target.fs.path("file.jpg"), mime=True)
        "image/jpg"

    Currently does not implement ``python-magic`` :class:`Magic` invocation behaviour.

    Resources:
        - https://github.com/ahupp/python-magic/blob/master/magic/__init__.py
        - https://github.com/file/file/tree/master/magic
        - https://freedesktop.org/wiki/Specifications/shared-mime-info-spec/
        - https://gitlab.freedesktop.org/xdg/shared-mime-info/-/blob/master/data/freedesktop.org.xml.in
    """

    @staticmethod
    def detect(buf: BinaryIO, suffix: str | None = None, *, mime: bool = False) -> MagicResult:
        """Searches ``mimetypes.MAP`` for the given bytes."""
        if suffix is not None and not isinstance(suffix, str):
            raise TypeError("Provided suffix is not a string")

        res_attr = "type" if mime else "name"
        for index, offset, magic in mimetypes.MAP:
            buf.seek(offset)
            if buf.read(len(magic)) == magic:
                return mimetypes.TYPES[index][res_attr]

        if suffix:
            for index, patterns in mimetypes.PATTERNS:
                if suffix.endswith(patterns):
                    return mimetypes.TYPES[index][res_attr]

        return None


def from_file(path: Path, *, mime: bool = False) -> MagicResult:
    """Detect file type from a :class:`Path` instance."""

    if not isinstance(path, Path):
        raise TypeError("Provided path is not a Path instance")

    with path.open("rb") as fh:
        return from_descriptor(fh, path.suffix, mime=mime)


def from_entry(entry: FilesystemEntry, *, mime: bool = False) -> MagicResult:
    """Detect file type from a :class:`FilesystemEntry` instance."""

    if not isinstance(entry, FilesystemEntry):
        raise TypeError("Provided entry is not a FilesystemEntry instance")

    with entry.open() as fh:
        return from_descriptor(fh, Path(entry.name).suffix, mime=mime)


def from_descriptor(fh: BinaryIO, suffix: str | None = None, *, mime: bool = False) -> MagicResult:
    """Detect file type from a file descriptor or handle."""

    if not hasattr(fh, "read") or not hasattr(fh, "seek"):
        raise TypeError("Provided fh does not have a read or seek method")

    return Magic().detect(fh, suffix, mime=mime)


# Convenience alias, not present in python-magic.
from_fh = from_descriptor


def from_buffer(buf: bytes, suffix: str | None = None, *, mime: bool = False) -> MagicResult:
    """Detect file type from provided bytes or buffer of bytes."""

    if not isinstance(buf, bytes):
        raise TypeError("Provided buf is not bytes")

    return from_descriptor(BytesIO(buf), suffix, mime=mime)



from __future__ import annotations

from unittest.mock import Mock

import pytest
from dissect.btrfs.btrfs import INode
from dissect.btrfs.c_btrfs import c_btrfs

from dissect.target.filesystems.btrfs import BtrfsFilesystemEntry


@pytest.mark.parametrize(
    ("sector_size", "filesize", "expected_blocks"),
    [
        (0x1000, 0x343, 0x8),
        (0x1000, 0x1000, 0x8),
        (0x1000, 0x1001, 0x10),
    ],
    ids=["lower", "equal", "greater"],
)
def test_stat_information_file_blocksize(sector_size: int, filesize: int, expected_blocks: int) -> None:
    entry = INode(Mock(), 42)
    entry.btrfs = Mock(sector_size=sector_size)
    timestamp = c_btrfs.btrfs_timespec()
    entry.inode = c_btrfs.btrfs_inode_item(
        mode=0o777,
        nlink=0,
        uid=1000,
        gid=1000,
        size=filesize,
        atime=timestamp,
        ctime=timestamp,
        otime=timestamp,
        mtime=timestamp,
    )

    fs_entry = BtrfsFilesystemEntry(Mock(), "some/path", entry)

    stat_info = fs_entry.lstat()

    assert stat_info.st_mode == 0o777
    assert stat_info.st_ino == 42
    assert stat_info.st_dev == 0
    assert stat_info.st_nlink == 0
    assert stat_info.st_uid == 1000
    assert stat_info.st_gid == 1000
    assert stat_info.st_size == filesize

    assert stat_info.st_atime == 0.0
    assert stat_info.st_atime_ns == 0
    assert stat_info.st_mtime == 0.0
    assert stat_info.st_mtime_ns == 0
    assert stat_info.st_ctime == 0.0
    assert stat_info.st_ctime_ns == 0
    assert stat_info.st_birthtime_ns == 0
    assert stat_info.st_birthtime_ns == 0

    assert stat_info.st_blksize == sector_size
    assert stat_info.st_blocks == expected_blocks


def test_stat_directory() -> None:
    """Using btrfs stat information as a base."""
    entry = INode(Mock(), 42, type=c_btrfs.BTRFS_FT_DIR)
    entry.btrfs = Mock(sector_size=0x1000)
    timestamp = c_btrfs.btrfs_timespec()
    entry.inode = c_btrfs.btrfs_inode_item(
        mode=0o777,
        nlink=0,
        uid=1000,
        gid=1000,
        size=0x1000,
        atime=timestamp,
        ctime=timestamp,
        otime=timestamp,
        mtime=timestamp,
    )

    fs_entry = BtrfsFilesystemEntry(Mock(), "some/path", entry)

    stat_info = fs_entry.lstat()

    assert stat_info.st_blocks == 0



from datetime import datetime, timezone

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.windows.everything._plugin import EverythingPlugin
from dissect.target.target import Target
from tests._utils import absolute_path

DIRECTORY = "directory"
FILE = "file"


def test_everything_1_7_20_ntfs_only(target_win: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_NTFS_ONLY.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = []
    idx = 0
    for idx, record in enumerate(target_win.everything.locate()):
        if idx not in [0, 86828]:
            continue
        records.append(record)

    length = idx + 1
    assert length == 126828

    assert records[0].type == DIRECTORY
    assert records[0].path == "C:\\$Recycle.Bin"
    assert records[0].date_modified == datetime(2023, 12, 16, 9, 8, 42, 726189, tzinfo=timezone.utc)
    assert records[1].type == FILE
    assert records[1].path == "C:\\Windows\\System32\\msidntld.dll"
    assert records[1].date_modified == datetime(2021, 5, 8, 8, 14, 34, 642525, tzinfo=timezone.utc)


def test_everything_1_7_20_ntfs_and_folder(target_win: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_NTFS_AND_FOLDER.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = []

    idx = 0
    for idx, record in enumerate(target_win.everything.locate()):
        if idx not in [0, 1, 94783]:
            continue
        records.append(record)

    length = idx + 1
    assert length == 127042

    assert records[0].type == DIRECTORY
    assert records[0].path == "C:\\$Recycle.Bin"
    assert records[0].date_modified == datetime(2023, 12, 16, 9, 8, 42, 726189, tzinfo=timezone.utc)

    assert records[1].type == DIRECTORY
    assert records[1].path == "E:\\$RECYCLE.BIN"
    assert records[1].date_modified == datetime(2024, 1, 27, 9, 54, 32, tzinfo=timezone.utc)

    assert records[2].type == FILE
    assert records[2].path == "E:\\potato.txt"
    assert records[2].date_modified == datetime(2024, 1, 27, 9, 54, 32, tzinfo=timezone.utc)


def test_everything_1_7_20_ntfs_and_refs(target_win: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_NTFS_AND_REFS.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = []
    idx = 0
    for idx, record in enumerate(target_win.everything.locate()):
        if idx not in [0, 1, 94937, 108011]:
            continue

        records.append(record)

    length = idx + 1
    assert length == 127227

    assert records[0].type == DIRECTORY
    assert records[0].path == "C:\\$Recycle.Bin"
    assert records[0].date_modified == datetime(2023, 12, 16, 9, 8, 42, 726189, tzinfo=timezone.utc)

    assert records[1].type == DIRECTORY
    assert records[1].path == "E:\\$RECYCLE.BIN"
    assert records[1].date_modified == datetime(2024, 1, 27, 21, 37, 6, 675776, tzinfo=timezone.utc)

    assert records[2].type == FILE
    assert records[2].path == "E:\\potato.txt"
    assert records[2].date_modified == datetime(2024, 1, 27, 21, 37, 13, 425810, tzinfo=timezone.utc)

    assert records[3].type == FILE
    assert records[3].path == "E:\\test.txt"
    assert records[3].date_modified == datetime(2024, 1, 27, 21, 56, 57, 597593, tzinfo=timezone.utc)


def test_everything_1_7_20_refs_with_disabled_ntfs(target_win: Target, fs_win: VirtualFilesystem) -> None:
    # Just making sure that disabling a drive doesn't mess up parsing
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_REFS_WITH_DISABLED_NTFS.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = list(target_win.everything.locate())
    assert len(records) == 7
    record = records[0]
    assert record.type == DIRECTORY
    assert record.path == "E:\\$RECYCLE.BIN"
    assert record.date_modified == datetime(2024, 1, 27, 21, 37, 6, 675776, tzinfo=timezone.utc)

    record = records[5]
    assert record.type == FILE
    assert record.path == "E:\\potato.txt"
    assert record.date_modified == datetime(2024, 1, 27, 21, 37, 13, 425810, tzinfo=timezone.utc)


def test_everything_1_7_20_refs_include_only(target_win: Target, fs_win: VirtualFilesystem) -> None:
    # Just making sure that including specific subfolders doesn't mess up parsing
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_REFS_INCLUDE_ONLY.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = list(target_win.everything.locate())
    assert len(records) == 4
    record = records[0]
    assert record.type == DIRECTORY
    assert record.path == "E:\\test"
    assert record.date_modified == datetime(2024, 1, 27, 22, 6, 9, 597820, tzinfo=timezone.utc)

    record = records[3]
    assert record.type == FILE
    assert record.path == "E:\\test\\test.txt"
    assert record.date_modified == datetime(2024, 1, 27, 21, 56, 57, 597593, tzinfo=timezone.utc)


def test_everything_1_7_20_efu(target_win: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_FILE_LIST.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = list(target_win.everything.locate())
    assert len(records) == 5
    record = records[0]
    assert record.type == DIRECTORY
    assert record.path == "E:"
    assert record.date_modified is None

    record = records[1]
    assert record.type == DIRECTORY
    assert record.path == "E:\\test"
    assert record.date_modified == datetime(2024, 1, 27, 22, 6, 9, 597820, tzinfo=timezone.utc)


def test_everything_1_7_17_ntfs_refs_efu_folder(target_win: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.17/Everything_NTFS_REFS_EFU_FOLDER.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = []
    idx = 0
    for idx, record in enumerate(target_win.everything.locate()):
        if idx not in [0, 1]:
            continue
        records.append(record)

    length = idx + 1
    assert length == 127577

    assert records[0].type == DIRECTORY
    assert records[0].path == "C:\\$Recycle.Bin"
    assert records[0].date_modified == datetime(2023, 12, 16, 9, 8, 42, 726189, tzinfo=timezone.utc)
    assert records[0].date_created is None

    assert records[1].type == DIRECTORY
    assert records[1].path == "C:\\$WinREAgent"
    assert records[1].date_modified == datetime(2024, 1, 26, 20, 49, 55, 41714, tzinfo=timezone.utc)


@pytest.mark.parametrize(
    "map_path",
    [
        "\\Program Files\\Everything\\Everything.db",
        "\\Program Files\\Everything\\Everything.COMPNAME.USERNAME.db",
        "\\Program Files (x86)\\Everything\\Everything.db",
        "\\Users\\John\\AppData\\Local\\Everything\\Everything.db",
        "\\Users\\John\\AppData\\Local\\Everything\\Everything.COMPNAME.John.db",
    ],
)
def test_everything_path_mapper(target_win_users: Target, fs_win: VirtualFilesystem, map_path: str) -> None:
    fs_win.map_file(
        map_path,
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_NTFS_ONLY.db.gz"),
        compression="gzip",
    )
    assert EverythingPlugin(target_win_users).check_compatible() is None



from __future__ import annotations

import tempfile
from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch
from uuid import UUID

import pytest
from flow.record.fieldtypes import posix_path

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix._os import UnixPlugin, parse_fstab
from dissect.target.target import Target

if TYPE_CHECKING:
    from pathlib import Path

FSTAB_CONTENT = """
# /etc/fstab: static file system information.
#
# Use 'blkid' to print the universally unique identifier for a
# device; this may be used with UUID= as a more robust way to name devices
# that works even if disks are added and removed. See fstab(5).
#
# <file system> <mount point>   <type>  <options>       <dump>  <pass>

proc                                      /proc        proc    nodev,noexec,nosuid 0    0

UUID=563f929e-ab4b-4741-b0f4-e3843c9a7a6a /            ext4    defaults,discard    0    0

UUID=5d1f1508-069b-4274-9bfa-ae2bf7ffb5e0 /home        ext4    defaults            0    2

UUID=be0afdc6-10bb-4744-a71c-02e0e2812160 none         swap    sw                  0    0

/dev/mapper/vgubuntu-swap_1               none         swap    sw                  0   0

UUID=28a25297-9825-4f87-ac41-f9c20cd5db4f /boot        ext4    defaults            0    2

UUID=F631-BECA                            /boot/efi    vfat    defaults,discard,umask=0077   0    0

/dev/disk/cloud/azure_resource-part1      /mnt         auto    defaults,nofail,x-systemd.requires=cloud-init.service,comment=cloudconfig   0   2

/dev/mapper/vg--main-lv--var              /var         auto    default             0    2

/dev/vg-main/lv-data                      /data        auto    default             0    2

/dev/disk/by-uuid/af0b9707-0945-499a-a37d-4da23d8dd245 /moredata auto default      0    2

LABEL=foo                                 /foo         auto    default             0    2

localhost:/home/user/nfstest              /mnt/nfs     nfs     ro                  0    0
"""  # noqa


def test_parse_fstab(tmp_path: Path) -> None:
    with tempfile.NamedTemporaryFile(dir=tmp_path, delete=False) as tf:
        tf.write(FSTAB_CONTENT.encode("ascii"))
        tf.close()

        fs = VirtualFilesystem()
        fs.map_file("/etc/fstab", tf.name)

        records = list(parse_fstab(fs.path("/etc/fstab")))

    # 11 input records minus
    #   2 unsupported mount devices (proc, /dev/disk/cloud/azure_resource-part1)
    #   2 swap partitions
    #   1 root partition
    # = 6 expected results

    assert set(records) == {
        (UUID("5d1f1508-069b-4274-9bfa-ae2bf7ffb5e0"), None, "/home", "ext4", "defaults"),
        (UUID("28a25297-9825-4f87-ac41-f9c20cd5db4f"), None, "/boot", "ext4", "defaults"),
        (UUID("af0b9707-0945-499a-a37d-4da23d8dd245"), None, "/moredata", "auto", "default"),
        ("F631-BECA", None, "/boot/efi", "vfat", "defaults,discard,umask=0077"),
        (None, "vg--main-lv--var", "/var", "auto", "default"),
        (None, "vg--main-lv--data", "/data", "auto", "default"),
        (None, "foo", "/foo", "auto", "default"),
        ("localhost", "/home/user/nfstest", "/mnt/nfs", "nfs", "ro"),
    }


def test_mount_volume_name_regression(fs_unix: VirtualFilesystem) -> None:
    mock_fs = Mock()
    mock_vol = Mock()

    mock_vol.name = "test-volume"

    mock_fs.__type__ = "ext"
    mock_fs.extfs.volume_name = "ext-volume"
    mock_fs.volume = mock_vol
    mock_fs.exists.return_value = False

    for expected_volume_name in ["test-volume", "ext-volume"]:
        with patch(
            "dissect.target.plugins.os.unix._os.parse_fstab",
            return_value=[(None, expected_volume_name, "/mnt", "auto", "default")],
        ):
            target = Target()
            target.filesystems.add(mock_fs)
            UnixPlugin.create(target, fs_unix)

            assert target.fs.mounts["/mnt"] == mock_fs


@pytest.mark.parametrize(
    ("hostname_content", "hosts_content", "expected_hostname", "expected_domain"),
    [
        (b"", b"", "localhost", None),
        (b"", b"127.0.0.1 mydomain", "mydomain", "mydomain"),
        (b"", b"127.0.0.1 localhost", "localhost", None),
        (b"hostname", b"127.0.0.1 localhost\n::1 ip6-localhost", "hostname", None),
        (b"hostname.example.internal", b"127.0.0.1 localhost\n::1 ip6-localhost", "hostname", "example.internal"),
        (b"myhost", b"", "myhost", None),
        (b"myhost.mydomain", b"", "myhost", "mydomain"),
        (b"myhost", b"127.0.0.1 mydomain", "myhost", "mydomain"),
        (b"myhost.mydomain", b"127.0.0.1 localhost", "myhost", "mydomain"),
        (b"myhost.localhost", b"127.0.0.1 mydomain", "myhost", "mydomain"),
        (b"myhost.mycoolerdomain", b"127.0.0.1 mydomain", "myhost", "mycoolerdomain"),
        (b"localhost.mycoolerdomain", b"127.0.0.1 mydomain", "localhost", "mycoolerdomain"),
        (b"localhost.mycoolerdomain", b"127.0.0.1 localhost", "localhost", "mycoolerdomain"),
    ],
)
def test_parse_domain(
    target_unix: Target,
    fs_unix: VirtualFilesystem,
    hostname_content: bytes,
    hosts_content: bytes,
    expected_domain: str,
    expected_hostname: str,
) -> None:
    fs_unix.map_file_fh("/etc/hostname", BytesIO(hostname_content))
    fs_unix.map_file_fh("/etc/hosts", BytesIO(hosts_content))
    target_unix.add_plugin(UnixPlugin)

    assert target_unix.hostname == expected_hostname, (
        f"Expected hostname {expected_hostname!r} but got {target_unix.hostname!r}"
    )
    assert target_unix.domain == expected_domain, f"Expected domain {expected_domain!r} but got {target_unix.domain!r}"


@pytest.mark.parametrize(
    ("path", "expected_hostname", "expected_domain", "file_content"),
    [
        ("/etc/hostname", "myhost", "mydomain.com", b"myhost.mydomain.com"),
        ("/etc/HOSTNAME", "myhost", "mydomain.com", b"myhost.mydomain.com"),
        (
            "/etc/sysconfig/network",
            "myhost",
            "mydomain.com",
            b"NETWORKING=NO\nHOSTNAME=myhost.mydomain.com\nGATEWAY=192.168.1.1",
        ),
        ("/etc/hostname", "myhost", None, b"myhost"),
        ("/etc/sysconfig/network", "myhost", None, b"NETWORKING=NO\nHOSTNAME=myhost\nGATEWAY=192.168.1.1"),
        ("/not_a_valid_hostname_path", None, None, b""),
        ("/etc/hostname", None, None, b""),
        ("/etc/sysconfig/network", None, None, b""),
        ("/proc/sys/kernel/hostname", "myhost", None, b"myhost"),
    ],
)
def test_parse_hostname_string(
    target_unix: Target,
    fs_unix: VirtualFilesystem,
    path: Path,
    expected_hostname: str | None,
    expected_domain: str | None,
    file_content: str,
) -> None:
    fs_unix.map_file_fh(path, BytesIO(file_content))

    hostname, domain = target_unix._os._parse_hostname_string()

    assert hostname == expected_hostname, f"Expected hostname {expected_hostname!r} but got {hostname!r}"
    assert domain == expected_domain, f"Expected domain {expected_domain!r} but got {domain!r}"


def test_users(target_unix_users: Target) -> None:
    users = list(target_unix_users.users())

    assert len(users) == 3

    assert users[0].name == "root"
    assert users[0].uid == 0
    assert users[0].gid == 0
    assert users[0].home == posix_path("/root")
    assert users[0].shell == "/bin/bash"

    assert users[1].name == "user"
    assert users[1].uid == 1000
    assert users[1].gid == 1000
    assert users[1].home == posix_path("/home/user")
    assert users[1].shell == "/bin/bash"

    assert users[2].name == "+@ngtest"
    assert users[2].uid is None
    assert users[2].gid is None
    assert users[2].home == posix_path("")
    assert users[2].shell == ""


@pytest.mark.parametrize(
    ("expected_arch", "elf_buf"),
    [
        # https://launchpad.net/ubuntu/+source/coreutils/9.4-3.1ubuntu1
        ("x86_64-unix", "7f454c4602010100000000000000000003003e0001000000a06d000000000000"),  # amd64
        ("aarch64-unix", "7f454c460201010000000000000000000300b70001000000405e000000000000"),  # arm64
        ("aarch32-unix", "7f454c4601010100000000000000000003002800010000001d40000034000000"),  # armhf
        ("x86_32-unix", "7f454c460101010000000000000000000300030001000000e042000034000000"),  # i386
        ("powerpc64-unix", "7f454c4602010100000000000000000003001500010000007470000000000000"),  # ppc64el
        ("riscv64-unix", "7f454c460201010000000000000000000300f30001000000685a000000000000"),  # riscv64
    ],
)
def test_architecture(target_unix: Target, fs_unix: VirtualFilesystem, expected_arch: str, elf_buf: str) -> None:
    """Test if we correctly parse unix architecture."""
    fs_unix.map_file_fh("/bin/ls", BytesIO(bytes.fromhex(elf_buf)))
    assert target_unix.architecture == expected_arch



from __future__ import annotations

import hashlib
import io
import posixpath
import shutil
import struct
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

try:
    from Crypto.Cipher import AES

    HAS_PYCRYPTODOME = True
except ImportError:
    HAS_PYCRYPTODOME = False

from dissect.util.stream import AlignedStream, RelativeStream, ZlibStream

from dissect.target.exceptions import LoaderError
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.helpers import keychain
from dissect.target.loader import Loader
from dissect.target.plugins.os.unix.linux.android._os import AndroidPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target

DIRECTORY_MAPPING = {
    "a": "/data/app/{id}",
    "f": "/data/data/{id}/files",
    "db": "/data/data/{id}/databases",
    "ef": "/storage/emulated/0/Android/data/{id}",
    "sp": "/data/data/{id}/shared_preferences",
    "r": "/data/data/{id}",
    "obb": "/storage/emulated/0/Android/obb/{id}",
}


class AndroidBackupLoader(Loader):
    """Load Android backup files.

    References:
        - http://fileformats.archiveteam.org/wiki/Android_ADB_Backup
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        self.ab = AndroidBackup(path.open("rb"))

        if self.ab.encrypted:
            for key in keychain.get_keys_for_provider("ab") + keychain.get_keys_without_provider():
                if key.key_type == keychain.KeyType.PASSPHRASE:
                    try:
                        self.ab.unlock(key.value)
                        break
                    except ValueError:
                        continue
            else:
                raise LoaderError(f"Missing password for encrypted Android Backup: {path}")

    @staticmethod
    def detect(path: Path) -> bool:
        if path.suffix.lower() == ".ab":
            return True

        if path.is_file():
            # The file extension can be chosen freely so let's also test for the file magic
            with path.open("rb") as fh:
                if fh.read(15) == b"ANDROID BACKUP\n":
                    return True

        return False

    def map(self, target: Target) -> None:
        if self.ab.compressed or self.ab.encrypted:
            if self.ab.compressed and not self.ab.encrypted:
                word = "compressed"
            elif self.ab.encrypted and not self.ab.compressed:
                word = "encrypted"
            else:
                word = "compressed and encrypted"

            target.log.warning(
                "Backup file is %s, consider unwrapping with `python -m dissect.target.loaders.ab <path/to/backup.ab>`",
                word,
            )

        vfs = VirtualFilesystem(case_sensitive=False)

        fs = TarFilesystem(self.ab.open())
        for app in fs.path("/apps").iterdir():
            for subdir in app.iterdir():
                if subdir.name not in DIRECTORY_MAPPING:
                    continue

                path = DIRECTORY_MAPPING[subdir.name].format(id=app.name)

                # TODO: Remove once we move towards "directory entries"
                entry = subdir.get()
                entry.name = posixpath.basename(path)

                vfs.map_file_entry(path, entry)

        target.filesystems.add(vfs)
        target._os_plugin = AndroidPlugin.create(target, vfs)


class AndroidBackup:
    def __init__(self, fh: BinaryIO):
        self.fh = fh

        size = fh.seek(0, io.SEEK_END)
        fh.seek(0)

        # Don't readline() straight away as we may be reading something other than a backup file
        magic = fh.read(15)
        if magic != b"ANDROID BACKUP\n":
            raise ValueError("Not a valid Android Backup file")

        self.version = int(fh.read(2)[:1])
        self.compressed = bool(int(fh.read(2)[:1]))

        self.encrypted = False
        self.unlocked = True
        self.encryption = fh.readline().strip().decode()

        if self.encryption != "none":
            self.encrypted = True
            self.unlocked = False
            self._user_salt = bytes.fromhex(fh.readline().strip().decode())
            self._ck_salt = bytes.fromhex(fh.readline().strip().decode())
            self._rounds = int(fh.readline().strip())
            self._user_iv = bytes.fromhex(fh.readline().strip().decode())
            self._master_key = bytes.fromhex(fh.readline().strip().decode())

            self._mk = None
            self._iv = None

        self._data_offset = fh.tell()
        self.size = size - self._data_offset

    def unlock(self, password: str) -> None:
        if not self.encrypted:
            raise ValueError("Android Backup is not encrypted")

        self._mk, self._iv = self._decrypt_mk(password)
        self.unlocked = True

    def _decrypt_mk(self, password: str) -> tuple[bytes, bytes]:
        user_key = hashlib.pbkdf2_hmac("sha1", password.encode(), self._user_salt, self._rounds, 32)

        blob = AES.new(user_key, AES.MODE_CBC, iv=self._user_iv).decrypt(self._master_key)
        blob = blob[: -blob[-1]]

        offset = 0
        iv_len = blob[offset]
        offset += 1

        iv = blob[offset : offset + iv_len]
        offset += iv_len

        if offset >= len(blob):
            raise ValueError("Invalid password: bogus IV length")

        mk_len = blob[offset]
        offset += 1

        if offset >= len(blob):
            raise ValueError("Invalid password: bogus master key length")

        mk = blob[offset : offset + mk_len]
        offset += mk_len

        if offset >= len(blob):
            raise ValueError("Invalid password: bogus IV length")

        checksum_len = blob[offset]
        offset += 1

        if offset >= len(blob):
            raise ValueError("Invalid password: bogus checksum length")

        checksum = blob[offset : offset + checksum_len]

        ck_mk = _encode_bytes(mk) if self.version >= 2 else mk
        our_checksum = hashlib.pbkdf2_hmac("sha1", ck_mk, self._ck_salt, self._rounds, 32)
        if our_checksum != checksum:
            # Try reverse encoding for good measure
            ck_mk = mk if self.version >= 2 else _encode_bytes(mk)
            our_checksum = hashlib.pbkdf2_hmac("sha1", ck_mk, self._ck_salt, self._rounds, 32)

        if our_checksum != checksum:
            raise ValueError("Invalid password: master key checksum does not match")

        return mk, iv

    def open(self) -> BinaryIO:
        fh = RelativeStream(self.fh, self._data_offset)

        if self.encrypted:
            if not self.unlocked:
                raise ValueError("Missing password for encrypted Android Backup")
            fh = CipherStream(fh, self._mk, self._iv, self.size)

        if self.compressed:
            fh = ZlibStream(fh)

        return fh


class CipherStream(AlignedStream):
    """Transparently AES-CBC decrypted stream."""

    def __init__(self, fh: BinaryIO, key: bytes, iv: bytes, size: int):
        self._fh = fh

        self._key = key
        self._iv = iv
        self._cipher = None
        self._cipher_offset = 0
        self._reset_cipher()

        super().__init__(size)

    def _reset_cipher(self) -> None:
        self._cipher = AES.new(self._key, AES.MODE_CBC, iv=self._iv)
        self._cipher_offset = 0

    def _seek_cipher(self, offset: int) -> None:
        """CBC is dependent on previous blocks so to seek the cipher, decrypt and discard to the wanted offset."""
        if offset < self._cipher_offset:
            self._reset_cipher()
            self._fh.seek(0)

        while self._cipher_offset < offset:
            read_size = min(offset - self._cipher_offset, self.align)
            self._cipher.decrypt(self._fh.read(read_size))
            self._cipher_offset += read_size

    def _read(self, offset: int, length: int) -> bytes:
        self._seek_cipher(offset)

        data = self._cipher.decrypt(self._fh.read(length))
        if offset + length >= self.size:
            # Remove padding
            data = data[: -data[-1]]

        self._cipher_offset += len(data)

        return data


def _encode_bytes(buf: bytes) -> bytes:
    # Emulate byte[] -> char[] -> utf8 byte[] casting
    return struct.pack(">32h", *struct.unpack(">32b", buf)).decode("utf-16-be").encode("utf-8")


def main() -> None:
    import argparse

    parser = argparse.ArgumentParser(description="Android Backup file unwrapper")
    parser.add_argument("path", type=Path, help="source path")
    parser.add_argument("-p", "--password", help="encryption password")
    parser.add_argument("-t", "--tar", action="store_true", help="write a tar file instead of a plain Android Backup")
    parser.add_argument("-o", "--output", type=Path, help="output path")
    args = parser.parse_args()

    if not args.path.is_file():
        parser.exit("source path does not exist or is not a file")

    ext = ".tar" if args.tar else ".plain.ab"
    if args.output is None:
        output = args.path.with_suffix(ext)
    elif args.output.is_dir():
        output = args.output.joinpath(args.path.name).with_suffix(ext)
    else:
        output = args.output

    if output.exists():
        parser.exit(f"output path already exists: {output}")

    print(f"unwrapping {args.path} -> {output}")
    with args.path.open("rb") as fh:
        ab = AndroidBackup(fh)

        if ab.encrypted:
            if not args.password:
                parser.exit("missing password for encrypted Android Backup")
            ab.unlock(args.password)

        with ab.open() as fhab, output.open("wb") as fhout:
            if not args.tar:
                fhout.write(b"ANDROID BACKUP\n")  # header
                fhout.write(b"5\n")  # version
                fhout.write(b"0\n")  # compressed
                fhout.write(b"none\n")  # encryption

            shutil.copyfileobj(fhab, fhout, 1024 * 1024 * 64)


if __name__ == "__main__":
    main()



from __future__ import annotations

import codecs
from typing import TYPE_CHECKING

from dissect.cstruct import cstruct
from dissect.util.ts import wintimestamp

from dissect.target.exceptions import RegistryValueNotFoundError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.helpers.shell_folder_ids import DESCRIPTIONS
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

userassist_def = """
struct VERSION5_ENTRY {
    char padding0[4];
    uint32 number_of_executions;
    uint32 application_focus_count;
    uint32 application_focus_duration;
    char padding1[44];
    uint64 timestamp;
    char padding2[4];
};

struct VERSION3_ENTRY {
    uint32  session_id;
    uint32  number_of_executions;
    uint64  timestamp;
};
"""
c_userassist = cstruct().load(userassist_def)

UserAssistRecordDescriptor = create_extended_descriptor(
    [
        RegistryRecordDescriptorExtension,
        UserRecordDescriptorExtension,
    ]
)

UserAssistRecord = UserAssistRecordDescriptor(
    "windows/registry/userassist",
    [
        ("datetime", "ts"),
        ("path", "path"),
        ("uint32", "number_of_executions"),
        ("uint32", "application_focus_count"),
        ("uint32", "application_focus_duration"),
    ],
)


class UserAssistPlugin(Plugin):
    """UserAssist plugin."""

    KEY = "HKEY_CURRENT_USER\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\UserAssist"

    def check_compatible(self) -> None:
        if not len(list(self.target.registry.keys(self.KEY))) > 0:
            raise UnsupportedPluginError("No UserAssist key found")

    @export(record=UserAssistRecord)
    def userassist(self) -> Iterator[UserAssistRecord]:
        """Return the UserAssist information for each user.

        The UserAssist registry keys contain information about programs that were recently executed on the system.
        Programs launch via the commandline are not registered within these registry keys.

        References:
            - https://www.magnetforensics.com/blog/artifact-profile-userassist/
            - https://www.aldeid.com/wiki/Windows-userassist-keys

        Yields UserAssistRecords with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The entry timestamp.
            path (path): The entry path.
            number_of_executions (int): The number of executions for this entry.
            application_focus_count (int): The number of focus acount for this entry.
            application_focus_duration (int): The duration of focus for this entry.
        """
        for reg in self.target.registry.keys(self.KEY):
            user = self.target.registry.get_user(reg)
            for subkey in reg.subkeys():
                try:
                    version = subkey.value("Version").value
                except RegistryValueNotFoundError:
                    version = None

                for count in subkey.subkeys():
                    for entry in count.values():
                        timestamp = 0
                        number_of_executions = None
                        application_focus_count = None
                        application_focus_duration = None

                        if version == 5 and len(entry.value) == 72:
                            data = c_userassist.VERSION5_ENTRY(entry.value)
                            timestamp = data.timestamp
                            number_of_executions = data.number_of_executions
                            application_focus_count = data.application_focus_count
                            application_focus_duration = data.application_focus_duration
                        elif version == 3 and len(entry.value) == 16:
                            data = c_userassist.VERSION3_ENTRY(entry.value)
                            timestamp = data.timestamp
                            number_of_executions = data.number_of_executions
                        elif (version == 3 and len(entry.value) == 8) or (version is None and len(entry.value) == 16):
                            # Unknown format?
                            pass
                        else:
                            self.target.log.debug(
                                "Invalid userassist value of length %d: %r", len(entry.value), entry.value
                            )
                            continue

                        value = codecs.decode(entry.name, "rot-13")
                        parts = value.split("/")

                        try:
                            value = value.replace(parts[0], DESCRIPTIONS[parts[0][1:-1].lower()])
                        except KeyError:
                            pass

                        yield UserAssistRecord(
                            ts=wintimestamp(timestamp),
                            path=self.target.fs.path(value),
                            number_of_executions=number_of_executions,
                            application_focus_count=application_focus_count,
                            application_focus_duration=application_focus_duration,
                            _target=self.target,
                            _user=user,
                            _key=count,
                        )



from __future__ import annotations

import contextlib
import hashlib
from io import BytesIO
from pathlib import Path

from dissect.target.helpers import hashutil
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.helpers.logging import get_logger

try:
    import yara

    HAS_YARA = True

except ImportError:
    HAS_YARA = False

from typing import TYPE_CHECKING

from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Iterator


log = get_logger(__name__)

YaraMatchRecord = TargetRecordDescriptor(
    "filesystem/yara/match",
    [
        ("datetime", "ts_mtime"),
        ("path", "path"),
        ("string", "rule"),
        ("string[]", "matches"),
        ("string[]", "tags"),
        ("digest", "digest"),
        ("string", "namespace"),
    ],
)

DEFAULT_MAX_SCAN_SIZE = 10 * 1024 * 1024


class YaraPlugin(Plugin):
    """Plugin to scan files against a local YARA rules file."""

    def check_compatible(self) -> None:
        if not HAS_YARA:
            raise UnsupportedPluginError("Please install 'yara-python' to use the yara plugin.")

    @arg("-r", "--rules", required=True, nargs="*", help="path(s) to YARA rule file(s) or folder(s)")
    @arg("-p", "--path", default="/", help="path on target(s) to recursively scan")
    @arg("-m", "--max-size", type=int, default=DEFAULT_MAX_SCAN_SIZE, help="maximum file size in bytes to scan")
    @arg("-c", "--check", action="store_true", help="check if every YARA rule is valid")
    @arg("--no-decompress", action="store_true", help="do not automatically decompress files for extra scanning")
    @export(record=YaraMatchRecord)
    def yara(
        self,
        rules: list[str | Path],
        path: str = "/",
        max_size: int = DEFAULT_MAX_SCAN_SIZE,
        check: bool = False,
        no_decompress: bool = False,
    ) -> Iterator[YaraMatchRecord]:
        """Scan files inside the target up to a given maximum size with YARA rule file(s).

        Args:
            rules: ``list`` of strings or ``Path`` objects pointing to rule files to use.
            path: ``string`` of absolute target path to scan.
            max_size: Files larger than this size will not be scanned.
            check: Check if provided rules are valid, only compiles valid rules.
            no_decompress: Do not automatically decompress compressed files before scanning.

        Returns:
            Iterator yields ``YaraMatchRecord``.
        """

        compiled_rules = process_rules(rules, check)

        if not rules:
            self.target.log.error("No working rules found in '%s'", ",".join(rules))
            return

        if hasattr(compiled_rules, "warnings") and (num_warns := len(compiled_rules.warnings)) > 0:
            self.target.log.warning("YARA generated %s warnings while compiling rules", num_warns)
            for warning in compiled_rules.warnings:
                self.target.log.info(warning)

        self.target.log.warning("Will not scan files larger than %s MB", max_size // 1024 // 1024)

        if no_decompress:
            self.target.log.info("Will not automatically check in compressed files")
        else:
            self.target.log.info(
                "Will check in compressed files automatically (at most %s MB after decompression)",
                max_size // 1024 // 1024,
            )

        for _, _, files in self.target.fs.walk_ext(path):
            for file in files:
                fhandles = []
                try:
                    if (file_size := file.stat().st_size) > max_size:
                        self.target.log.info("Not scanning file of %s MB: '%s'", (file_size // 1024 // 1024), file)
                        continue

                    fh_original = file.open()
                    fhandles.append(fh_original)
                    if not no_decompress:
                        fh_decompress = open_decompress(fileobj=fh_original, mode="rb")
                        if fh_decompress is not fh_original:
                            fhandles.append(fh_decompress)

                    for fh in fhandles:
                        buf = fh.read(max_size + 1)
                        if len(buf) > max_size:
                            self.target.log.warning(
                                "Decompressed '%s' is larger than the maximum scan size, scanning up to %s bytes",
                                file,
                                max_size,
                            )

                        # ensure that the original file handle is reset for compressed reads
                        fh_original.seek(0)

                        for match in compiled_rules.match(data=buf):
                            string_matches: list[str] = []
                            for string in match.strings:
                                string_matches.extend(f"{string}={instance}" for instance in string.instances)

                            yield YaraMatchRecord(
                                ts_mtime=file.stat().st_mtime,
                                path=self.target.fs.path(file.path),
                                rule=match.rule,
                                matches=string_matches,
                                tags=match.tags,
                                digest=hashutil.common(BytesIO(buf)),
                                namespace=match.namespace,
                                _target=self.target,
                            )

                except FileNotFoundError:
                    continue
                except RuntimeWarning as e:
                    self.target.log.warning("Runtime warning while scanning file '%s': %s", file, e)
                except Exception as e:
                    self.target.log.error("Exception scanning file '%s'", file)  # noqa: TRY400
                    self.target.log.debug("", exc_info=e)
                finally:
                    for fh in reversed(fhandles):
                        with contextlib.suppress(Exception):
                            fh.close()


def process_rules(paths: list[str | Path], check: bool = False) -> yara.Rules | None:
    """Generate compiled YARA rules from the given path(s).

    Provide path to one (compiled) YARA file or directory containing YARA files.

    Args:
        paths: Path to file(s) or folder(s) containing YARA files.
        check: Attempt to compile every rule file before appending to rules.

    Returns:
        Compiled YARA rules or None.
    """
    files = set()
    compiled_rules = None

    for rules_path in paths:
        if isinstance(rules_path, str):
            rules_path = Path(rules_path)

        if not rules_path.exists():
            log.warning("File %s does not exist!", rules_path)
            continue

        if rules_path.is_dir():
            for file in rules_path.rglob("*"):
                if not file.is_file():
                    continue
                files.add(file)
        else:
            files.add(rules_path)

    for file in set(files):
        with file.open("rb") as fh:
            magic = fh.read(4)

        if magic == b"YARA":
            if len(files) > 1:
                log.error("Providing multiple compiled YARA files is not supported. Did not add %s", file)
                continue
            log.info("Adding single compiled YARA file %s", file)
            compiled_rules = compile_yara(file, is_compiled=True)
            break

        if check and not is_valid_yara({"check_namespace": file}):
            log.warning("File %s contains invalid rule(s)!", file)
            files.remove(file)
            continue

    if files and not compiled_rules:
        try:
            compiled_rules = compile_yara({hashlib.md5(file.as_posix().encode()).hexdigest(): file for file in files})
        except yara.Error as e:
            log.error("Failed to compile YARA file(s): %s", e)  # noqa: TRY400

    return compiled_rules


def compile_yara(files: dict[str, Path] | Path, is_compiled: bool = False) -> yara.Rules | None:
    """Compile or load the given YARA file(s) to rules."""
    if is_compiled and isinstance(files, Path):
        return yara.load(files.as_posix())
    return yara.compile(filepaths={ns: Path(path).as_posix() for ns, path in files.items()})


def is_valid_yara(files: dict[str, Path] | Path, is_compiled: bool = False) -> bool:
    """Determine if the given YARA file(s) compile without errors or warnings."""
    try:
        compile_yara(files, is_compiled)
    except (yara.SyntaxError, yara.WarningError, yara.Error) as e:
        log.debug("Rule file(s) '%s' invalid: %s", files, e)
        return False
    else:
        return True



from __future__ import annotations

from collections.abc import Callable, Generator, Iterator

from flow.record import GroupedRecord

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

OSInfoRecord = TargetRecordDescriptor(
    "generic/osinfo",
    [
        ("string", "name"),
        ("string", "value"),
    ],
)


class OSInfoPlugin(Plugin):
    """Convenience plugin that wraps _os.* functions in records."""

    def check_compatible(self) -> None:
        if not self.target._os_plugin:
            raise UnsupportedPluginError("No operating system detected on target")

    @export(record=OSInfoRecord)
    def osinfo(self) -> Iterator[OSInfoRecord | GroupedRecord]:
        """Yield grouped records with target OS info."""
        for os_func in self.target._os.__exports__:
            value = getattr(self.target._os, os_func)
            record = OSInfoRecord(name=os_func, value=None, _target=self.target)
            if isinstance(value, Callable) and isinstance(subrecords := value(), Generator):
                try:
                    yield GroupedRecord("generic/osinfo/grouped", [record, *list(subrecords)])
                except Exception:
                    # Ignore exceptions triggered by functions
                    # that cannot be executed in this context
                    continue
            else:
                record.value = value
                yield record



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import call, create_autospec, mock_open, patch

import pytest

from dissect.target.containers.raw import RawContainer
from dissect.target.filesystems.dir import DirectoryFilesystem
from dissect.target.loaders.local import (
    LINUX_DEV_DIR,
    LINUX_DRIVE_REGEX,
    LocalLoader,
    _add_disk_as_raw_container_to_target,
    _get_windows_drive_volumes,
    map_linux_drives,
)
from dissect.target.target import Target, TargetLogAdapter

if TYPE_CHECKING:
    from collections.abc import Callable


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``LocalLoader`` when opening a ``Target``."""
    for path in ["local", "local?some=query"]:
        with patch("dissect.target.target.Target.apply"):
            target = opener(path)
            assert isinstance(target._loader, LocalLoader)
            assert target.path == Path("local")

            if "?" in path:
                assert target.path_query == {"some": "query"}


def test_skip_emulated_drive() -> None:
    """Test that we skip emulated drives on Windows."""

    class Dummy:
        def __init__(self, data: dict):
            self.__dict__ = data

    dummy = Dummy(
        {"NumberOfDiskExtents": 1, "Extents": [Dummy({"DiskNumber": "999", "StartingOffset": 0, "ExtentLength": 0})]}
    )

    with (
        patch("builtins.open", create=True),
        patch("ctypes.windll", create=True),
        patch("ctypes.windll.kernel32.GetDriveTypeA", create=True, return_value=3),
        patch("dissect.util.stream.BufferedStream", create=True),
        patch("dissect.target.loaders.local._read_drive_letters", create=True, return_value=[b"Z:"]),
        patch("dissect.target.volume.Volume", create=True),
        patch("dissect.target.target.TargetLogAdapter", create=True) as mock_log,
        patch("dissect.target.loaders.local._windows_get_volume_disk_extents", create=True) as mock_extents,
    ):
        mock_extents.return_value = dummy

        for _ in _get_windows_drive_volumes(mock_log):
            pass

        assert (
            call.debug(
                "Skipped drive %d from %s, not a physical drive (could be emulation or ram disk)", "999", "\\\\.\\z:"
            )
            in mock_log.mock_calls
        )


def test_add_disk_as_raw_container_to_target() -> None:
    """Test that we add a disk as a raw container to the target."""
    # Does it attempt to open the file and pass a raw container?
    mock = mock_open()
    drive = Path("/xdev/fake")
    t = Target()

    with (
        patch("pathlib.Path.open", mock),
        patch.object(t.disks, "add") as mock_method,
    ):
        _add_disk_as_raw_container_to_target(drive, t)

        assert isinstance(mock_method.call_args[0][0], RawContainer) is True
        mock.assert_called_with("rb")


def test_add_disk_as_raw_container_to_target_skip_fail() -> None:
    """Test that we skip adding a disk as a raw container if it fails to open."""
    # Does it emit a warning instead of raising an exception?
    mock = mock_open()
    mock.side_effect = IOError
    drive = Path("/xdev/fake")
    t = Target()

    with (
        patch.object(TargetLogAdapter, "warning") as mock_warning,
        patch.object(TargetLogAdapter, "debug") as mock_debug,
        patch("pathlib.Path.open", mock),
    ):
        _add_disk_as_raw_container_to_target(drive, t)

        assert mock_warning.call_args[0] == ("Unable to open drive: %s, skipped", drive)
        assert isinstance(mock_debug.call_args[1]["exc_info"], OSError) is True
        mock.assert_called_with("rb")


def test_map_linux_drives(tmp_path: Path) -> None:
    """Test that we correctly map Linux drives to the target."""
    mock_drive = LINUX_DEV_DIR.joinpath("sda")
    mock_dev_dir = create_autospec(Path)
    mock_dev_dir.iterdir.return_value = iter([mock_drive])

    t = Target()

    with (
        patch("dissect.target.loaders.local.LINUX_DEV_DIR", mock_dev_dir),
        patch(
            "dissect.target.loaders.local._add_disk_as_raw_container_to_target",
            autospec=True,
        ) as mock_add_raw_disks,
        patch("dissect.target.loaders.local.VOLATILE_LINUX_PATHS", [tmp_path]),
        patch.object(t.filesystems, "add", autospec=True) as mock_add_fs,
        patch.object(t.fs, "mount", autospec=True) as mock_mount,
    ):
        map_linux_drives(t)

        mock_add_raw_disks.assert_called_with(mock_drive, t)

        mock_add_fs.assert_called()
        dir_fs = mock_add_fs.call_args[0][0]
        assert isinstance(dir_fs, DirectoryFilesystem)
        assert dir_fs.base_path == tmp_path

        mock_mount.assert_called_with(str(tmp_path), dir_fs)


@pytest.mark.parametrize(
    ("drive_path", "expected"),
    [
        (Path("/dev/fd0"), True),  # Floppy
        (Path("/dev/fd1"), True),  # Floppy
        (Path("/dev/fd2"), True),  # Floppy
        (Path("/dev/sda"), True),  # SCSI
        (Path("/dev/sdb"), True),  # SCSI
        (Path("/dev/sdc"), True),  # SCSI
        (Path("/dev/sda1"), False),  # Partition
        (Path("/dev/nvme0n1"), True),  # NVMe Disk
        (Path("/dev/nvme0n1p1"), False),  # Partition
        (Path("/dev/hda"), True),  # IDE-Controller
        (Path("/dev/hdb"), True),  # IDE-Controller
        (Path("/dev/hda1"), False),  # Partition
        (Path("/dev/vda"), True),  # Virtual hard disk
        (Path("/dev/vdb"), True),  # Virtual hard disk
        (Path("/dev/vda1"), False),  # Partition
        (Path("/dev/sr0"), False),  # Compact Disc
        (Path("/dev/scd0"), False),  # Compact Disc
        (Path("/dev/tty11"), False),  # Not a disk
        (Path("/dev/xsdaa"), False),  # Fake
        (Path("/dev/xhdaa"), False),  # Fake
    ],
)
def test_linux_drive_regex(drive_path: Path, expected: bool) -> None:
    """Test that we correctly match Linux drive paths."""
    assert (LINUX_DRIVE_REGEX.match(drive_path.name) is not None) == expected



from __future__ import annotations

import argparse
import hashlib
import subprocess
import sys
from datetime import datetime, timezone
from pathlib import Path

import defusedxml.ElementTree as ET

from dissect.target.tools.utils.cli import catch_sigpipe

TEMPLATE = """\
# This file is generated by target-build-magic
# Do not edit manually, for overrides see 'overrides.py'
#
# Generated on:         {timestamp}
# Source SHA1 hash:     {hash}
# Number of types:      {count}
# References:           http://www.freedesktop.org/standards/shared-mime-info
#                       https://gitlab.freedesktop.org/xdg/shared-mime-info/-/blob/master/data/freedesktop.org.xml.in

TYPES = {types}
"""

IGNORED_TYPES = (
    # Conflicts with ELF
    "application/x-iso9660-appimage",
    "application/vnd.appimage",
    "application/x-core",
    # Conflicts with application/zip
    "application/x-karbon",
    "application/x-kchart",
    "application/x-kformula",
    "application/x-kivio",
    "application/x-kontour",
    "application/x-kpresenter",
    "application/x-krita",
    "application/x-kspread",
    "application/x-kword",
    # Conflicts with TIFF
    "image/x-3ds",
    # Conflicts with MZ
    "application/x-msdownload",
    # Conflicts with TAR
    "text/x-devicetree-source",
    # Conflicts with DEB
    "application/x-archive",
    # Conflicts with XML
    "application/x-quicktime-media-link",
    # Conflicts with OGG
    "application/annodex",
    # Conflicts with ZLIB
    "text/x-todo-txt",
    # Conflicts with PNG
    "image/apng",
    # Too broad definition
    "video/mp2t",
)

IGNORED_SUB_TYPES = (
    "application/zip",
    "application/xml",
    "application/json",
    "application/annodex",
)


def parse_freedesktop_xml(path: Path) -> str:
    """Attempt to parse a FreeDesktop shared-mime-info document.

    Does not yet parse nested magic values.
    """

    ns = {"mime": "http://www.freedesktop.org/standards/shared-mime-info"}
    types = []

    content = path.read_bytes()
    hash = hashlib.sha1(content).hexdigest()

    magics_seen = set()

    for mime_type in ET.fromstring(content.decode()).findall("mime:mime-type", ns):
        mime_type_name = mime_type.get("type")

        if (sub_class := mime_type.find("mime:sub-class-of", ns)) is not None and sub_class.get(
            "type"
        ) in IGNORED_SUB_TYPES:
            continue

        if mime_type_name in IGNORED_TYPES:
            continue

        comment = mime_type.find("mime:comment", ns).text
        magics = []

        for magic_match in mime_type.findall("mime:magic/mime:match", ns):
            value = magic_match.get("value")
            type = magic_match.get("type")

            if not value:
                raise ValueError(f"Missing value in {magic_match!r}")

            if not type:
                raise ValueError(f"Missing type in {magic_match!r}")

            try:
                value = convert_value(type, value)
            except Exception as e:
                print(f"Failed to convert {type!r} with value {value!r} for type {mime_type_name!r}: {e!s}")
                continue

            # Calculate offset
            if ":" in (offset := magic_match.get("offset", "0")):
                offset = offset.split(":")[0]
            try:
                offset = int(offset)
            except ValueError:
                pass

            if (offset, value) in magics_seen:
                print(f"Duplicate magic value for type {mime_type_name!r}: offset={offset!r} value={value!r}")
                continue

            if len(value) < 2:
                print(
                    f"Magic value too short to be usable for type {mime_type_name!r}: offset={offset!r} value={value!r}"
                )
                continue

            magics_seen.add((offset, value))
            magics.append({"offset": offset, "value": value})

        patterns = [g.get("pattern") for g in mime_type.findall("mime:glob", ns)]

        if not patterns and not magics:
            print(f"No usable definitions for {mime_type_name!r}")
            continue

        types.append(
            {
                "type": mime_type_name,
                "name": comment,
                "pattern": patterns,
                "magic": magics,
            }
        )

    print(f"Parsed {len(types)} types")

    return TEMPLATE.format(timestamp=datetime.now(tz=timezone.utc), count=len(types), hash=hash, types=types)


def convert_value(type: str, value: str) -> bytes:
    """Convert a value to bytes."""

    # Regular string conversion to bytes
    # e.g. "PK\\003\\004"       -> b"PK\x03\x04"
    # e.g. "\\x00\\x01"         -> b"\x00\x01"
    # e.g. "\x89PNG\r\n\x1A\n"  -> b"\x89PNG\r\n\x1a\n"
    if type == "string":
        if "\\x" in value or "\\0" in value:
            return value.encode("latin1").decode("unicode-escape").encode("latin1")
        return value.encode("latin1")

    # Integer conversion
    # e.g. "0x9501"     -> b"\x95\x01"
    # e.g. "0xa1b2c3d4" -> b"\xa1\xb2\xc3\xd4"
    if type in ("byte", "little16", "little32", "big16", "big32", "host16", "host32"):
        return bytes.fromhex(value[2:] if value[:2] == "0x" else value)

    raise ValueError(f"Unknown type {type!r}")


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="target-build-magic",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )
    parser.add_argument("path", metavar="PATH", type=Path, help="path to freedesktop.org.xml.in file")
    parser.add_argument(
        "--output", "-o", type=Path, help="optional alternative output path for generated types document"
    )
    parser.add_argument("--format", action="store_true", default=False, help="format output using ruff")
    args = parser.parse_args()

    if not args.path.is_file():
        print(f"Provided file does not exist: {args.path}")
        return 1

    output = (args.output or Path(__file__).parents[1].joinpath("helpers/magic/mimetypes/freedesktop.py")).resolve()

    if not output.parent.is_dir():
        print(f"Output path parent directory {output.parent} does not exist")
        return 1

    try:
        content = parse_freedesktop_xml(args.path)
    except Exception as e:
        print(f"Exception while generating output: {e!s}")
        return 1

    try:
        output.write_text(content)
    except Exception as e:
        print(f"Exception while writing output to {output!s}: {e!s}")
        return 1

    if args.format:
        subprocess.run(["ruff", "format", output], capture_output=True)
        subprocess.run(["ruff", "check", "--fix", "--unsafe-fixes", output], capture_output=True)

    return 0


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

import binascii
from collections.abc import Iterator
from datetime import datetime
from enum import IntEnum
from io import BytesIO
from typing import TYPE_CHECKING

from dissect.cstruct import Structure, cstruct
from dissect.util.ts import wintimestamp

from dissect.target.exceptions import Error, RegistryError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

ShimcacheRecord = TargetRecordDescriptor(
    "windows/shimcache",
    [
        ("datetime", "last_modified"),
        ("string", "name"),
        ("varint", "index"),
        ("path", "path"),
    ],
)

shim_def = """
struct NT61_HEADER {
    uint32 magic;
    uint32 num_entries;
};

struct NT61_64_ENTRY {
    uint16 len;
    uint16 maxlen;
    uint32 _align;
    uint64 offset;
    uint64 ts;
    uint32 fileflags;
    uint32 flags;
    uint64 blobsize;
    uint64 bloboffset;
};

struct NT52_HEADER {
    uint32 magic;
    uint32 num_entries;
};

struct NT52_ENTRY_32 {
    uint16 len;
    uint16 maxlen;
    uint32 offset;
    uint64 ts;
    uint64 filesize;
};


struct NT52_ENTRY_64 {
    uint16 len;
    uint16 maxlen;
    uint32 _padding;
    uint64 offset;
    uint64 ts;
    uint64 filesize;
};

struct WIN81_ENTRY {
    uint32 magic;
    uint32 crc;
    uint32 len;
    char data[len];
};

struct WIN81_ENTRY_DATA {
    uint16 path_len;
    wchar path[path_len/2];
    uint16 pkg_len;
    wchar pkg[pkg_len/2];
    uint32 flags;
    uint32 a;
    uint64 ts;
    uint32 b;
};

struct WIN81_ENTRY_DATA_SINGLE {
    uint16 path_len;
    wchar path[path_len/2];
    uint32 flags;
};

struct WIN10_ENTRY {
    uint32 magic;
    uint32 crc;
    uint32 len;
    char data[len];
};

struct WIN10_ENTRY_DATA {
    uint16 path_len;
    wchar path[path_len/2];
    uint64 ts;
};
"""
c_shim = cstruct().load(shim_def)

MAGIC_NT61 = 0xBADC0FEE
MAGIC_NT52 = 0xBADC0FFE
MAGIC_WIN81 = 0x73743031
MAGIC_WIN10 = 0x73743031


class SHIMCACHE_WIN_TYPE(IntEnum):
    """Shimcache version mapping."""

    VERSION_WIN10_CREATORS = 0x1001
    VERSION_WIN10 = 0x1000
    VERSION_WIN81 = 0x0801
    VERSION_NT61 = 0x0601
    VERSION_NT52 = 0x0502

    VERSION_WIN81_NO_HEADER = 0x1002  # auto()


def win_10_path(ed: Structure) -> str:
    return ed.path


def win_8_path(ed: Structure) -> str:
    return ed.path if ed.path_len else ed.pkg


def nt52_entry_type(fh: bytes) -> Structure:
    entry = c_shim.NT52_ENTRY_32(fh)

    return c_shim.NT52_ENTRY_64 if entry.offset == 0 else c_shim.NT52_ENTRY_32


def nt61_entry_type(fh: bytes) -> Structure:
    return c_shim.NT61_64_ENTRY


TYPE_VARIATIONS = {
    SHIMCACHE_WIN_TYPE.VERSION_WIN10: {
        "headers": (c_shim.WIN10_ENTRY, c_shim.WIN10_ENTRY_DATA),
        "offset": 0x30,
        "path_finder": win_10_path,
    },
    SHIMCACHE_WIN_TYPE.VERSION_WIN10_CREATORS: {
        "headers": (c_shim.WIN10_ENTRY, c_shim.WIN10_ENTRY_DATA),
        "offset": 0x34,
        "path_finder": win_10_path,
    },
    SHIMCACHE_WIN_TYPE.VERSION_WIN81: {
        "headers": (c_shim.WIN81_ENTRY, c_shim.WIN81_ENTRY_DATA),
        "offset": 0x80,
        "path_finder": win_8_path,
    },
    SHIMCACHE_WIN_TYPE.VERSION_WIN81_NO_HEADER: {
        "headers": (c_shim.WIN81_ENTRY, c_shim.WIN81_ENTRY_DATA_SINGLE),
        "offset": 0x0,
        "path_finder": win_8_path,
    },
    SHIMCACHE_WIN_TYPE.VERSION_NT61: {
        "header": c_shim.NT61_HEADER,
        "header_function": nt61_entry_type,
        "offset": 0x80,
    },
    SHIMCACHE_WIN_TYPE.VERSION_NT52: {
        "header": c_shim.NT52_HEADER,
        "header_function": nt52_entry_type,
        "offset": 0x8,
    },
}


class CRCMismatchException(Error):
    pass


ShimCacheGeneratorType = CRCMismatchException | tuple[datetime | None, str]


class ShimCache:
    def __init__(self, fh: BytesIO, ntversion: str, noheader: bool = False):
        self.fh = fh
        self.ntversion = ntversion
        self.noheader = noheader

        self.version = self.identify()

    def __iter__(self) -> Iterator[ShimCacheGeneratorType]:
        if self.version not in list(SHIMCACHE_WIN_TYPE):
            raise NotImplementedError

        arguments = TYPE_VARIATIONS.get(self.version)

        if self.version in (
            SHIMCACHE_WIN_TYPE.VERSION_NT61,
            SHIMCACHE_WIN_TYPE.VERSION_NT52,
        ):
            shimcache_iterator = self.iter_nt
        else:
            shimcache_iterator = self.iter_win_8_plus

        return shimcache_iterator(**arguments)

    def identify(self) -> SHIMCACHE_WIN_TYPE:
        """Identify which SHIMCACHE version to use."""
        self.fh.seek(0)
        d = self.fh.read(0x100)
        magic = c_shim.uint32(d[:4])

        if magic == MAGIC_NT52:
            return SHIMCACHE_WIN_TYPE.VERSION_NT52

        if magic == MAGIC_NT61:
            return SHIMCACHE_WIN_TYPE.VERSION_NT61

        if magic == MAGIC_WIN81 and self.ntversion == "6.3":
            self.noheader = True
            return SHIMCACHE_WIN_TYPE.VERSION_WIN81_NO_HEADER

        if len(d) >= 0x84 and c_shim.uint32(d[0x80:0x84]) == MAGIC_WIN81:
            return SHIMCACHE_WIN_TYPE.VERSION_WIN81

        if len(d) >= 0x34 and c_shim.uint32(d[0x30:0x34]) == MAGIC_WIN10:
            return SHIMCACHE_WIN_TYPE.VERSION_WIN10

        if len(d) >= 0x38 and c_shim.uint32(d[0x34:0x38]) == MAGIC_WIN10:
            return SHIMCACHE_WIN_TYPE.VERSION_WIN10_CREATORS

        if self.ntversion == "6.3":
            if self.noheader:
                return SHIMCACHE_WIN_TYPE.VERSION_WIN81_NO_HEADER
            return SHIMCACHE_WIN_TYPE.VERSION_WIN81

        raise NotImplementedError

    def iter_win_8_plus(
        self, headers: tuple[Structure, Structure], offset: int, path_finder: Callable
    ) -> Iterator[ShimCacheGeneratorType]:
        entry_header, data_header = headers

        self.fh.seek(offset)
        while True:
            try:
                entry = entry_header(self.fh)
            except EOFError:
                break

            if binascii.crc32(entry.data) & 0xFFFFFFFF != entry.crc:
                yield CRCMismatchException(message=f"offset={self.fh.tell() - len(entry)}")
                break

            ed = data_header(entry.data)
            path = path_finder(ed)

            yield wintimestamp(ed.ts) if hasattr(ed, "ts") else None, path

    def iter_nt(self, header: Structure, offset: int, header_function: Callable) -> Iterator[tuple[datetime, str]]:
        self.fh.seek(0)

        header = header(self.fh)
        entry_header = header_function(self.fh)

        self.fh.seek(offset)

        for _ in range(header.num_entries):
            pos = self.fh.tell()
            entry = entry_header(self.fh)

            self.fh.seek(entry.offset)
            path = self.fh.read(entry.len)

            try:
                path = path.decode("utf-16-le")
            except UnicodeDecodeError:
                break

            yield wintimestamp(entry.ts), path

            self.fh.seek(pos + len(entry_header))


class ShimcachePlugin(Plugin):
    """Shimcache plugin."""

    KEYS = (
        "HKLM\\SYSTEM\\CurrentControlSet\\Control\\Session Manager\\AppCompatCache",
        "HKLM\\SYSTEM\\CurrentControlSet\\Control\\Session Manager\\AppCompatibility",
    )

    def check_compatible(self) -> None:
        if not len(list(self.target.registry.keys(self.KEYS))) > 0:
            raise UnsupportedPluginError("Could not find shimcache")

    @export(record=ShimcacheRecord)
    def shimcache(self) -> Iterator[ShimcacheRecord]:
        """Return the shimcache.

        The ShimCache or AppCompatCache stores registry keys related to properties from older Windows versions for
        compatibility purposes. Since it contains information about files such as the last
        modified date and the file size, it can be useful in forensic investigations.

        References:
            - https://www.andreafortuna.org/2017/10/16/amcache-and-shimcache-in-forensic-analysis/

        Yields ShimcacheRecords with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            last_modified (datetime): The last modified date.
            name (string): The value name.
            index (varint): The index of the entry.
            path (uri): The parsed path.
        """
        for key in self.target.registry.keys(self.KEYS):
            for value_name in ("AppCompatCache", "CacheMainSdb"):
                try:
                    data = key.value(value_name).value
                except RegistryError:
                    continue

                try:
                    cache = ShimCache(BytesIO(data), self.target.ntversion, value_name != "AppCompatCache")
                except NotImplementedError:
                    self.target.log.warning("Not implemented ShimCache version: %s %s", key, value_name)
                    continue
                except EOFError:
                    self.target.log.warning("Error parsing ShimCache entry: %s %s", key, value_name)
                    continue

                yield from self._get_records(value_name, cache)

    def _get_records(self, name: str, cache: Iterator[ShimCacheGeneratorType]) -> Iterator[ShimcacheRecord]:
        for index, item in enumerate(cache):
            if isinstance(item, CRCMismatchException):
                self.target.log.warning("A CRC mismatch occured for entry: %s", item)
                continue

            ts, file_path = item

            yield ShimcacheRecord(
                last_modified=ts,
                name=name,
                index=index,
                path=self.target.resolve(file_path),
                _target=self.target,
            )



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualHive, VirtualKey
from dissect.target.plugins.apps.productivity.winrar import WinRarPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_winrar_windows(target_win_users: Target, hive_hku: VirtualHive) -> None:
    """Test if we detect WinRAR dialog artifacts."""

    key_path = "Software\\WinRAR\\DialogEditHistory\\ArcName"
    key = VirtualKey(hive_hku, key_path)
    key.add_value("0", "example.rar")
    key.timestamp = datetime(2025, 12, 31, 12, 34, 0, tzinfo=timezone.utc)
    hive_hku.map_key(key_path, key)

    key_path = "Software\\WinRAR\\DialogEditHistory\\ExtrPath"
    key = VirtualKey(hive_hku, key_path)
    key.add_value("0", "C:\\Users\\Administrator\\Desktop\\example-rar")
    key.timestamp = datetime(2025, 12, 31, 13, 37, 0, tzinfo=timezone.utc)
    hive_hku.map_key(key_path, key)

    target_win_users.add_plugin(WinRarPlugin)

    results = list(target_win_users.winrar())
    assert len(results) == 2

    assert results[0]._desc.name == "application/productivity/winrar"
    assert results[0].ts == datetime(2025, 12, 31, 12, 34, 0, tzinfo=timezone.utc)
    assert results[0].path == "example.rar"

    assert results[1]._desc.name == "application/productivity/winrar"
    assert results[1].ts == datetime(2025, 12, 31, 13, 37, 0, tzinfo=timezone.utc)
    assert results[1].path == "C:\\Users\\Administrator\\Desktop\\example-rar"



from __future__ import annotations

import importlib
import sys
from typing import TYPE_CHECKING

from dissect.target.helpers import keychain
from dissect.target.plugins.os.windows.dpapi.dpapi import DPAPIPlugin
from dissect.target.plugins.os.windows.dpapi.keyprovider.keychain import KeychainKeyProviderPlugin

if TYPE_CHECKING:
    import pytest

    from dissect.target.target import Target


def test_dpapi_keyprovider_keychain(target_win: Target) -> None:
    """test if we yield provided keychain items correctly."""
    keychain.register_key(
        key_type=keychain.KeyType.PASSPHRASE,
        value="password1",
        identifier=None,
        provider="user",
    )
    keychain.register_key(
        key_type=keychain.KeyType.PASSPHRASE,
        value="password2",
        identifier=None,
        provider=None,
    )

    target_win.add_plugin(DPAPIPlugin, check_compatible=False)
    target_win.add_plugin(KeychainKeyProviderPlugin)

    keys = list(target_win.dpapi.keyprovider.keychain())

    assert keys == [
        ("dpapi.keyprovider.keychain", "password1"),
        ("dpapi.keyprovider.keychain", "password2"),
    ]


def test_env_keychain(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test if we can read keychain items from environment variables."""
    KEYCHAIN_MODULE = "dissect.target.helpers.keychain"

    # Set environment variable before module import
    monkeypatch.delitem(sys.modules, KEYCHAIN_MODULE, raising=False)
    monkeypatch.setenv("DISSECT_KEYCHAIN_VALUE", "envtestpass")

    fresh_keychain = importlib.import_module(KEYCHAIN_MODULE)

    keys = fresh_keychain.get_all_keys()
    # There should be at least one key with value 'envtestpass' and is_wildcard True
    assert any(k.value == "envtestpass" and k.is_wildcard for k in keys)

    # Monkeypatch will roll back the original keychain module



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.volume.md.md import MD, MDPhysicalDisk, find_super_block

from dissect.target.volume import LogicalVolumeSystem, Volume

if TYPE_CHECKING:
    from collections.abc import Iterator
    from uuid import UUID

    from typing_extensions import Self


class MdVolumeSystem(LogicalVolumeSystem):
    __type__ = "md"

    def __init__(self, fh: BinaryIO | list[BinaryIO] | None, *args, **kwargs):
        self.md = MD(fh)
        super().__init__(fh, *args, **kwargs)

    @classmethod
    def open_all(cls, volumes: list[BinaryIO]) -> Iterator[Self]:
        devices: dict[UUID, list[MDPhysicalDisk]] = {}

        for vol in volumes:
            if not cls.detect_volume(vol):
                continue

            device = MDPhysicalDisk(vol)
            devices.setdefault(device.set_uuid, []).append(device)

        for devs in devices.values():
            try:
                yield cls(devs, disk=[dev.fh for dev in devs])
            except Exception:  # noqa: PERF203
                continue

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        vols = [fh] if not isinstance(fh, list) else fh
        return any(MdVolumeSystem.detect_volume(vol) for vol in vols)

    @staticmethod
    def _detect_volume(fh: BinaryIO) -> bool:
        offset, _, _ = find_super_block(fh)
        return offset is not None

    def _volumes(self) -> Iterator[Volume]:
        # MD only supports one configuration and virtual disk but doing this as a loop
        # makes it automatically safe for empty configurations
        for conf in self.md.configurations:
            for vd in conf.virtual_disks:
                fh = vd.open()
                yield Volume(fh, 1, None, vd.size, None, vd.name, vd.uuid, raw=self.md, disk=self.disk, vs=self)



from __future__ import annotations

import datetime
import gzip
from typing import TYPE_CHECKING, TextIO

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

STATUS_FILE_NAME = "/var/lib/dpkg/status"
LOG_FILES_GLOB = "/var/log/dpkg.log*"

STATUS_FIELD_MAPPINGS = {
    "Package": "name",
    "Status": "status",
    "Priority": "priority",
    "Section": "section",
    "Architecture": "arch",
    "Version": "version",
}

STATUS_FIELDS_TO_EXTRACT = STATUS_FIELD_MAPPINGS.keys()

DpkgPackageStatusRecord = TargetRecordDescriptor(
    "linux/debian/dpkg/package/status",
    [
        ("string", "name"),
        ("string", "status"),
        ("string", "priority"),
        ("string", "section"),
        ("string", "arch"),
        ("string", "version"),
    ],
)

DpkgPackageLogRecord = TargetRecordDescriptor(
    "linux/debian/dpkg/package/log",
    [
        ("datetime", "ts"),
        ("string", "name"),
        ("string", "operation"),
        ("string", "status"),
        ("string", "version_old"),
        ("string", "version"),
        ("string", "arch"),
    ],
)


class DpkgPlugin(Plugin):
    """Returns records for package details extracted from dpkg's status and log files."""

    __namespace__ = "dpkg"

    def check_compatible(self) -> None:
        log_files = list(self.target.fs.glob(LOG_FILES_GLOB))
        if not len(log_files) > 0 and not self.target.fs.path(STATUS_FILE_NAME).exists():
            raise UnsupportedPluginError("No DPKG files found")

    @export(record=DpkgPackageStatusRecord)
    def status(self) -> Iterator[DpkgPackageStatusRecord]:
        """Yield records for packages in dpkg's status database."""

        status_file_path = self.target.fs.path(STATUS_FILE_NAME)

        if not status_file_path.exists():
            return

        for block_lines in read_status_blocks(status_file_path.open("rt")):
            details = parse_status_block(block_lines)

            if not details or not details.get("Package"):
                continue

            record_fields = {
                STATUS_FIELD_MAPPINGS[field]: value
                for field, value in details.items()
                if field in STATUS_FIELD_MAPPINGS
            }

            yield DpkgPackageStatusRecord(_target=self.target, **record_fields)

    @export(record=DpkgPackageLogRecord)
    def log(self) -> Iterator[DpkgPackageLogRecord]:
        """Yield records for actions logged in dpkg's logs."""
        target_tz = self.target.datetime.tzinfo

        for log_file in self.target.fs.glob(LOG_FILES_GLOB):
            fh = self.target.fs.open(log_file)
            if log_file.lower().endswith(".gz"):
                fh = gzip.open(fh)  # noqa: SIM115

            for line in fh:
                line = line.decode("utf-8").strip()

                try:
                    parsed_line = parse_log_line(line, target_tz)
                except ValueError:
                    self.target.log.debug("Can not parse dpkg log line `%s`", line, exc_info=True)
                    continue

                if not parsed_line:
                    continue

                yield DpkgPackageLogRecord(_target=self.target, **parsed_line)


def read_status_blocks(fh: TextIO) -> Iterator[list[str]]:
    """Yield package status blocks read from ``fh`` text stream as the lists of lines."""
    block_lines = []
    for line in fh:
        line = line.strip()

        # Package details blocks are separated by an empty line
        if not line:
            if block_lines:
                yield block_lines
                block_lines = []
            continue

        block_lines.append(line)

    if block_lines:
        yield block_lines


def parse_status_block(block_lines: list[str]) -> dict[str, str]:
    """Parse package details block from dpkg status file."""
    result = {}
    for line in block_lines:
        field_name, _, value = line.partition(": ")
        if field_name in STATUS_FIELDS_TO_EXTRACT:
            result[field_name] = value.strip()
    return result


def parse_log_date_time(date_str: str, time_str: str, tzinfo: datetime.tzinfo = datetime.timezone.utc) -> datetime:
    return datetime.datetime.strptime(f"{date_str} {time_str}", "%Y-%m-%d %H:%M:%S").replace(tzinfo=tzinfo)


def parse_log_line(log_line: str, tzinfo: datetime.tzinfo = datetime.timezone.utc) -> dict[str, str]:
    """Parse dpkg log file line."""

    parts = log_line.split(" ")

    # Skip lines that are not about operations on packages
    if len(parts) != 6:
        return None

    result = {}
    log_date, log_time, operation = parts[:3]

    result = {
        "ts": parse_log_date_time(log_date, log_time, tzinfo=tzinfo),
        "operation": operation,
    }

    if operation == "status":
        # Example:
        # 2022-01-03 12:47:24 status unpacked python3.8:amd64 3.8.10-0ubuntu1~20.04.2

        status, package_arch, version = parts[3:]
        name, _, arch = package_arch.partition(":")
        result.update(
            {
                "name": name,
                "status": status,
                "arch": arch,
                "version": version,
            }
        )
    elif operation in ("install", "upgrade", "remove", "trigproc"):
        # Example:
        # 2022-01-03 12:47:41 install linux-modules-extra-5.11.0-43-generic:amd64 <none> 5.11.0-43.47~20.04.2
        package_arch, version_old, version = parts[3:]
        name, _, arch = package_arch.partition(":")
        version = None if version == "<none>" else version
        version_old = None if version_old == "<none>" else version_old
        result.update(
            {
                "name": name,
                "arch": arch,
                "version": version,
                "version_old": version_old,
            }
        )
    else:
        raise ValueError(f"Unrecognized operation `{operation}` in dpkg log file line: `{log_line}`")

    return result



#!/usr/bin/env python
from __future__ import annotations

import argparse
import logging
import pathlib
import shutil
import sys
from typing import TYPE_CHECKING

from dissect.target.exceptions import TargetError
from dissect.target.helpers.logging import get_logger
from dissect.target.target import Target
from dissect.target.tools.utils.cli import (
    catch_sigpipe,
    configure_generic_arguments,
    process_generic_arguments,
)
from dissect.target.tools.utils.fs import print_ls, print_stat

if TYPE_CHECKING:
    from dissect.target.helpers.fsutil import TargetPath


log = get_logger(__name__)
logging.lastResort = None
logging.raiseExceptions = False


def ls(t: Target, path: TargetPath, args: argparse.Namespace) -> None:
    if args.use_ctime and args.use_atime:
        log.error("Can't specify -c and -u at the same time")
        return
    if not path or not path.exists():
        return

    # Only output with colors if stdout is a tty
    use_colors = sys.stdout.buffer.isatty()

    print_ls(
        path,
        0,
        sys.stdout,
        args.l,
        args.human_readable,
        args.recursive,
        args.use_ctime,
        args.use_atime,
        use_colors,
    )


def cat(t: Target, path: TargetPath, args: argparse.Namespace) -> None:
    stdout = sys.stdout
    if hasattr(stdout, "buffer"):
        stdout = stdout.buffer
    shutil.copyfileobj(path.open(), stdout)


def walk(t: Target, path: TargetPath, args: argparse.Namespace) -> None:
    for e in path.rglob("*"):
        print(str(e))


def cp(t: Target, path: TargetPath, args: argparse.Namespace) -> None:
    output = pathlib.Path(args.output).expanduser().resolve()

    if path.is_file():
        _extract_path(path, output.joinpath(path.name))
    elif path.is_dir():
        for extract_path in path.rglob("*"):
            out_path = output.joinpath(str(extract_path.relative_to(path)))
            _extract_path(extract_path, out_path)
    else:
        print(f"[!] Failed, unsuported file type: {path}")


def stat(t: Target, path: TargetPath, args: argparse.Namespace) -> None:
    if not path or not path.exists():
        return
    print_stat(path, sys.stdout, args.dereference)


def _extract_path(path: TargetPath, output_path: pathlib.Path) -> None:
    print(f"{path} -> {output_path}")

    out_dir = output_path if path.is_dir() else output_path.parent

    try:
        if not out_dir.exists():
            out_dir.mkdir(parents=True)

        if path.is_file():
            with output_path.open("wb") as fh:
                shutil.copyfileobj(path.open(), fh)

    except Exception:
        print(f"[!] Failed: {path}")
        log.exception("Error extracting file: %s -> %s", path, output_path)


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="dissect.target",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )
    parser.add_argument("target", type=pathlib.Path, help="Target to load", metavar="TARGET")

    baseparser = argparse.ArgumentParser(add_help=False)
    baseparser.add_argument("path", help="path to perform an action on", metavar="PATH")

    subparsers = parser.add_subparsers(dest="subcommand", help="subcommands for performing various actions")
    parser_ls = subparsers.add_parser(
        "ls", help="Show a directory listing", parents=[baseparser], conflict_handler="resolve"
    )
    parser_ls.add_argument("-l", action="store_true")
    parser_ls.add_argument("-a", "--all", action="store_true")  # ignored but included for proper argument parsing
    parser_ls.add_argument("-h", "--human-readable", action="store_true")
    parser_ls.add_argument("-R", "--recursive", action="store_true", help="recursively list subdirectories encountered")
    parser_ls.add_argument(
        "-c", action="store_true", dest="use_ctime", help="show time when file status was last changed"
    )
    parser_ls.add_argument("-u", action="store_true", dest="use_atime", help="show time of last access")
    parser_ls.set_defaults(handler=ls)

    parser_cat = subparsers.add_parser("cat", help="dump file contents", parents=[baseparser])
    parser_cat.set_defaults(handler=cat)

    parser_stat = subparsers.add_parser("stat", help="display file status", parents=[baseparser])
    parser_stat.add_argument("-L", "--dereference", action="store_true")
    parser_stat.set_defaults(handler=stat)

    parser_find = subparsers.add_parser("walk", help="perform a walk", parents=[baseparser])
    parser_find.set_defaults(handler=walk)

    parser_cp = subparsers.add_parser(
        "cp",
        help="copy multiple files to a directory specified by --output",
        parents=[baseparser],
    )
    parser_cp.add_argument("-o", "--output", default=".", help="output directory")
    parser_cp.set_defaults(handler=cp)
    configure_generic_arguments(parser)

    args, _ = parser.parse_known_args()
    process_generic_arguments(parser, args)

    if args.subcommand is None:
        parser.error("No subcommand specified")

    try:
        target = Target.open(args.target)
    except TargetError as e:
        log.error(e)  # noqa: TRY400
        log.debug("", exc_info=e)
        return 1

    glob_path = str(args.path).lstrip("/")

    found = False
    for path in target.fs.path("/").glob(glob_path):
        args.handler(target, path, args)
        found = True

    if not found:
        print("[!] Path doesn't exist")
        return 1

    return 0


if __name__ == "__main__":
    main()
